import pytest
import os
import shutil
import tempfile
import textwrap
import pathlib
import yaml


@pytest.fixture
def config_factory():
    filename = tempfile.NamedTemporaryFile(suffix='.yml', delete=False).name

    def get_config(text_or_dict):
        with open(filename, 'w') as conf:
            if isinstance(text_or_dict, str):
                conf.write(textwrap.dedent(text_or_dict))
            elif isinstance(text_or_dict, dict):
                yaml.dump(text_or_dict, conf)
            else:
                raise ValueError(
                    'A text or dictionary is expected for config_factory.')
        return filename

    yield get_config
    os.remove(filename)


@pytest.fixture
def script_factory():
    filename = tempfile.NamedTemporaryFile(suffix='.sos', delete=False).name

    def get_script(text):
        with open(filename, 'w') as conf:
            conf.write(textwrap.dedent(text))
        return filename

    yield get_script
    os.remove(filename)


@pytest.fixture
def temp_factory():

    temp_fds = []
    temp_dirs = []

    def get_tempfiles(*args, **kwargs):
        content = kwargs.get('content', None)
        for names in args:
            if isinstance(names, str):
                names = [names]
            for name in names:
                if content is None:
                    pathlib.Path(name).touch()
                else:
                    with open(name, 'w') as tf:
                        tf.write(content)
                temp_fds.append(name)
        if 'dir' in kwargs:
            if isinstance(kwargs['dir'], str):
                dirs = [kwargs['dir']]
            else:
                dirs = kwargs['dir']
            for dir in dirs:
                if os.path.isdir(dir):
                    shutil.rmtree(dir)
                os.makedirs(dir, exist_ok=True)
                temp_dirs.append(dir)
        return temp_fds, temp_dirs

    yield get_tempfiles

    for temp_file in temp_fds:
        try:
            if os.path.isfile(temp_file):
                os.remove(temp_file)
        except Exception:
            pass

    for temp_dir in temp_dirs:
        try:
            if os.path.isdir(temp_dir):
                shutil.rmtree(temp_dir)
        finally:
            pass


@pytest.fixture
def clear_now_and_after():
    # This fixture takes one or more file names or directory names and will
    # remove them if they exist, and also after the tests are completed.
    temp_fds = []

    def clear_files_and_dirs():
        for temp_fd in temp_fds:
            try:
                if os.path.isfile(temp_fd):
                    os.remove(temp_fd)
                elif os.path.isdir(temp_fd):
                    shutil.rmtree(temp_fd)
            except Exception:
                pass

    def get_names(*args):
        for names in args:
            if isinstance(names, str):
                temp_fds.append(names)
            else:
                temp_fds.extend(names)

        clear_files_and_dirs()
        return temp_fds

    yield get_names

    clear_files_and_dirs()

@pytest.fixture
def sample_workflow():
    with open('sample_workflow.ipynb', 'w') as sn:
        sn.write(r'''{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {
    "kernel": "SoS"
   },
   "outputs": [],
   "source": [
    "# this is a test workflow"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "kernel": "SoS"
   },
   "source": [
    "This is a markdown cell"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "kernel": "R"
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "This is a cell with another kernel"
     ]
    }
   ],
   "source": [
    "cat('This is a cell with another kernel')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {
    "kernel": "SoS"
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "This is a scratch cell\n"
     ]
    }
   ],
   "source": [
    "print('This is a scratch cell')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "kernel": "SoS"
   },
   "outputs": [],
   "source": [
    "# this comment will be included but not shown in help message\n",
    "# because it is for the global\n",
    "[global]\n",
    "a = 1\n",
    "# this comment will become the comment for parameter b\n",
    "parameter: b=2\n",
    "parameter: c=3\n",
    "# this comment will become the comment for parameter d\n",
    "parameter: d='d'"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "kernel": "SoS"
   },
   "outputs": [],
   "source": [
    "# this comment will not be included in exported workflow\n",
    "# because it is not immediately before section\n",
    "\n",
    "# this is a section comment, will be displayed\n",
    "[default]\n",
    "print(f'Hello {a}')"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "SoS",
   "language": "sos",
   "name": "sos"
  },
  "language_info": {
   "codemirror_mode": "sos",
   "file_extension": ".sos",
   "mimetype": "text/x-sos",
   "name": "sos",
   "nbconvert_exporter": "sos_notebook.converter.SoS_Exporter",
   "pygments_lexer": "sos"
  },
  "sos": {
   "default_kernel": "SoS",
   "kernels": [
    [
     "R",
     "ir",
     "R",
     "#DCDCDA"
    ],
    [
     "SoS",
     "sos",
     "",
     ""
    ]
   ],
   "panel": {
    "displayed": true,
    "height": 0,
    "style": "side"
   },
   "version": "0.9.14.11"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
''')


@pytest.fixture
def fastq_files():
    if not os.path.isdir('data'):
        os.mkdir('data')
    with open('data/S20_R1.fastq', 'w') as R1:
        R1.write('''\
@ST-J00106:29:H3C5HBBXX:4:1101:1172:1000 1:N:0:TGACCA
CCTCAGTTCCTCAAACACAGTGAGTGAGTACGGACCTCAGGCTGTCTCTTCGCGTCACACCCGCTGTGAAATGGCAGATACAGCATCAGGGCAGACACTC
+
<<AAFKKKKKKKKKFKFKK7FKKKKKFK,7,<AFF<7,AAKKAK,FK7FKA,7A,,,,7,,,<(F<K7,F,,7<A<,,AA,<,<<,A7,F,(A7A7,AAK
@ST-J00106:29:H3C5HBBXX:4:1101:1375:1000 1:N:0:TGACCA
ATATGCATTCATATAAGTTCATGTCATGCAGCTGACAAATAGAATGATATAGCTCAACTCTACTCATGGATGCCGAACGATGTCTGTGGCTTGCACTGTG
+
AAFFAKKKFKKKKFKKKKKFKKKFKAFKKKKKKKKKKKKFKKKKAKKFKKKKF7FF<FKK<KKKFKKKKFKKKKKFAAKF,,AFFKK7,,7F7AFKK,FK
@ST-J00106:29:H3C5HBBXX:4:1101:1395:1000 1:N:0:TGACCA
CTGTTACCATGTTAATAGCCACACCCTTATGGCCAAACCATCCACCTGTATCCACAGAATTGGTCGAGGTGGATGGTTTGGCCATAAGGGTGTTGCTATG
+
AAAFFKKKKKFKKKKKKKKAKKKFFAFAFAKKKFKKKKFKKFKKAAAA7FFFAFKKKKA<AKKKKKFFK,,AAFFF<,,7F,,A,F<<F,F7F,FFKFAF
@ST-J00106:29:H3C5HBBXX:4:1101:1497:1000 1:N:0:TGACCA
GTTCGAGGGCAGGGTGAGGCCGCGGGGCTGAGCACCATGGCCGCTACCGGGACCACCAGGCCGTGCGCGTTTCCACGCTGTCTTTCTAGGATGCTCCC
+
AAFF,AFK<<KFAFKKKKKKFFKKFKKFKKKKFKFKKFKKKKKKKKKKKKKKFKKKKKKKKKKKFKKKKAF7FKKFF7FKKKKKKAAKF7FKA,<FA<
@ST-J00106:29:H3C5HBBXX:4:1101:1598:1000 1:N:0:TGACCA
ATAAAATAAATTTTGCATTCTAAAAATAAAAATGTAAAGAATATACAGCCATAGGCCCCAATTGTTTGCTTCTTCCAGGAACAGCATCTCCTCTTCTGA
+
AAAAAKAKAKAAFKKKK7<FAKKKKKKKKKKK7KFKKKKKFAFFAKKFKKKAK,FF7(<FKK<,<F<7,A<,FKK7FFFKFKFKKKFA,FAFKKKK7FA
@ST-J00106:29:H3C5HBBXX:4:1101:1619:1000 1:N:0:TGACCA
GATCAAGCGAGCTTTTGCCCTTCTGCTCCACGGGAGGTTTCTGTTCTCCTTGATCTCGCCTTAGGACAGCTCGCGGGGAGTCGAAGGACAAAAGCGCGTT
+
AA,FFKKKKKKK,<<AKKKKFFAKKKKAKKK7<FF(<,,,7,A,,7A<A,,,A,7,77,,,,,<F7,,,,,F,,A((,A7,,7,,,,,,AA7<,,,,A(,
@ST-J00106:29:H3C5HBBXX:4:1101:1923:1000 1:N:0:TGACCA
CATGAAAGTGAAGAACAAGGAAGCCCCAACCTACCTTGTTCTTCACTTTCATGCCAGACTACAACTGTACCCCACAAAGAGGTGGGGTACAGTTGTAGTC
+
AAFFFKKFKKKKKKKKKKKFKKKKKKKKKKKKKKKKKKKKKKKKKKKKKFKKKKAKKKKKKKKKKKKKKKAKKKK7FAAKK,7,A,AK7,<A7FKFFK7K
@ST-J00106:29:H3C5HBBXX:4:1101:1964:1000 1:N:0:TGACCA
AGGAAATATAGAGAATACCACAAAGAAACTCCTTGCAAAGAGCAACTCCAAGACCCATAATTGGCAGATTCACCAAGATGGAAATGAACGAAAAAATATT
+
AAFFAKKKKKKKKKKKKKKKKKKKKKKKKKKKAKKKKKKKKKKKKKFKKKKKKKAKKKFKKFK,FF7A,,AKFFFKKKAKKFFA,,FKFKKKKFKKA<AK
@ST-J00106:29:H3C5HBBXX:4:1101:2086:1000 1:N:0:TGACCA
TACCAACAAATGCAAAAACACAAACCAGTGGTAAATATGCAGACCAGCGATACTAGAAAGCAGCCACACAAACAAGACAGCATAATAACCAGCTAAGAAC
+
AAFA<FAKAAFKAKKKKKKF<AAFKKKK,A,A,FFFF,<<FKKF7FKK<A7AAAF,<FKKAF<,,A,A,AFF7A7<FA,7,A,,A,FKF,F,7,AFFKKK
@ST-J00106:29:H3C5HBBXX:4:1101:2187:1000 1:N:0:TGACCA
CCTTTCCCAGTCCACTAAATAAATGTTAATCTCCTTTGGCAACACCCTCACAGACACAACCGGGAAAAACACTTTGGATCCTTCGATCCAAT
+
AAAFFKKKKKFKKKKKKKKKKKKKKAKFFFKAFKKKKKKKKKKKKFKKKKKKKKKKKKKKKKK<FKKKKKKKKKKKKFFKAAAKKKKK<KK<
@ST-J00106:29:H3C5HBBXX:4:1101:2207:1000 1:N:0:TGACCA
GAAGAGCACACGTCTGAACTCCAGTCACTGACCAATCTCGTATGCCGTCTTCTGCTTGAAAAAAAAAAAAAAAAGAAGCAGGGAATGATTCCCTACCGG
+
AAFFFKKKKKKKFKKKKKKFKKKKKKKKFKKKKKK7,FFKKKKKKKKAAKKKKKKFFFKKKKKKKKKKA,,,,,,,,,,,,7,,7,,,,,,,,,,,,,,
@ST-J00106:29:H3C5HBBXX:4:1101:2289:1000 1:N:0:TGACCA
TATTGGATAGAATTCTTAATTCATTCTCTATGTTATCTTGAATTTCTTTGAGCTTCCTTAACACAGCTATTGTGAATTCCGATTCACAACACAAAAAC
+
AAFFFKKAKFKKKKKKKKKKKKKKKKKKAK7FFKA,AKKKAKKFKFA<K,<<FKFFFKKA,FFFFKKFKKKFKFA7FFFFA,<,,,FA,,7<A,,AFA
@ST-J00106:29:H3C5HBBXX:4:1101:2410:1000 1:N:0:TGACCA
TTTCTCTTCTGAGCTCCTAAAGTGCAGGGTGTTTTTCAGACTCAAAGGTCAAGGCCTTCTTAGAATAGGGCTGCATGTGTCTTTGTCCAGTGCATGTTGC
+
AAFF<<FFKKKKFKKKK<KKKKKKKFFKKKA<<A,,FKKKKKKKKKKKKFAFF,7<,,7,,7FF<FKFAKKAFFA,A7F,AFAFKFAKFFAF7F,77,,,
@ST-J00106:29:H3C5HBBXX:4:1101:2512:1000 1:N:0:TGACCA
TCAAGTCTTGAAGCTGGCCCAGGCTTACAGACCAGTTCGCTGAGGTAGCTACCACTCAGTAGAGGGCACTCCCTTTGTCCTGGGAGGCATAGCTGAGTAT
+
<AFFFKKKFKKKKKAKKKAFKKKKKKKKKKKKKFKAFKKKKKKKFFKFK7FAFKFFKKKKKKKFAKKKKKKKK,F,7AAFAKKFK7AAA<FA<7KK,7,<
@ST-J00106:29:H3C5HBBXX:4:1101:2593:1000 1:N:0:TGACCA
TAACTTTGTGTTTGAAGGAGGGGGGCAGGAGATCAGGACCAAGGGTTTGGCCAAGGCCAGATCTAGAAGACTCGGGCCATCCCATAGTTGATACCTT
+
AAFFFKKKKKKKFKKKKKKKKKKKKKKKKKFKAKKKKKKKKKKKKKKKKKKFAKFK<KKKKFFKKKKKKKKKKKKKKFFAFKKKKKKF,AFFFKKKA
@ST-J00106:29:H3C5HBBXX:4:1101:2613:1000 1:N:0:TGACCA
TATTTACAAACCCTGCCATAAAGTGCTCTTGAAAATTATGTCAAGGCCGGGCGCGGTGGCTCACGCCTGTAATCCCAGC
+
AAAFFKKKKKKKKKFFKKKKKKKKKKKKKFFFKKKFKKKKKKKKKKKKKKKKKKKKKKKKKKAFKKKKKFFKKKKKAKA
@ST-J00106:29:H3C5HBBXX:4:1101:2756:1000 1:N:0:TGACCA
CAGCCACGTGGCTTTAAGTACAATCTACATGACAACGACTCCTGTGGTAGCCAGCCTCCAAGATGGTCCCAGTGGCGCCTGCCTCTGTACCAGGGTTGGT
+
AAFFFKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKFKKKKKKKKF<<A<KKKKF<FFF(,FAFFK7<<FF,FAFKFF
@ST-J00106:29:H3C5HBBXX:4:1101:2918:1000 1:N:0:TGACCA
CCAGATGAGAAAATAAAGAAAGGGACATGGAAACAGGATGAGGAAAAACACTTTTTTTCTCATCTGGGCTATTCTTGAGGCAGGAGGAAAAAGAAGAGA
+
AAFFFFKKKKKKKKKKFKKKKKKKKKKKKA7FFKKKAFFKKKKKFKKKKKK7FK7<FAK7FKKKKKKFA7AAAAF,A,<FA7FF<AKK7FFFF7AKFKF
@ST-J00106:29:H3C5HBBXX:4:1101:3202:1000 1:N:0:TGACCA
TTTTTGAGATGGAGTCAGGCTCTGTCGCCCAGGCTGGAGTGCAGTGGTGCGATCTTGGCTCACTGCAAGCTCAGCCTCCCGGGTTCAGGCCA
+
AAFFFFAFKKKKKKKKKKKKFKKKKKKKKKKKKKAKKKKFKFKKKKKAKKKKAKKFKKKKKKKKKKKKKKFKKKKKK,FKK<F,FKKKKKKK
@ST-J00106:29:H3C5HBBXX:4:1101:3344:1000 1:N:0:TGACCA
AAAAACACACCAGAGAGAGACAGAGGTCCTAAAGATGCAGACCCCAGACCCTGAGACAGAGAGACCCTAGACCCACAGAGAAAGATACTTCACCACCCCC
+
AAFFFKKKKKFFKKKKKKKKKKKKKKKKKKKKKKKKKKKFFK7FFKKKKAF7FKKKAKKKKKKFFAFKKKKKKKKKFKKKKKKKKKKF<7AFFKKKKKKK
@ST-J00106:29:H3C5HBBXX:4:1101:3385:1000 1:N:0:TGACCA
TTCTTTCTCTTGCCTGATTGCCCTGGCTAGAACTTCCAATACTATGTTGAATGGGAGTGGTGAGAGAGGGCATCCTTGTTTTGTGCCGGGTTTCAAAGGG
+
AAFFAKKFKKFKKKKKKKFAFFKAFAKAFFKKKFAAAFKKKKFFKKKAFKKAFKKKKKKKFKKKK7F,FA(,F7AKKKKKFKKKK<KFF(FAAKF7AA77
@ST-J00106:29:H3C5HBBXX:4:1101:3425:1000 1:N:0:TGACCA
ACCAGAACCACTTTTTGGCTCCTTCTATAGTTTGAGTATGTCATCTTTATTTCTAAGTCTGTATGATATCAATTTTGCATTAGCACCTCGACTTTGA
+
AAFFFKKKKKKKKKKKKKKAKKKKKKKKKKKKKKKKFKKKKKKKKFKKKKFF7AKKFFKKFKKKKKFKKKKKKKFFKKKK7FAKKFKFAA,FKKKAA
@ST-J00106:29:H3C5HBBXX:4:1101:3446:1000 1:N:0:TGACCA
AGCCTAGACCTTTACCTGCATCTCCCTCAGGGCTAACTTCCCAAGACAGTGGCAGGTAAAGGTCTAGGCTCTCCACTTAGAC
+
AAFFFKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKFKKKKKFFKKKKKKKKKKKKKKKKKK
@ST-J00106:29:H3C5HBBXX:4:1101:3466:1000 1:N:0:TGACCA
AAGGTTGTAGAGTGACCTACAGGTGAACAGCTCTGCTCCTCCCAACTGTTCGACCACTGCTCTCACCTCCACACAATCCCAGTGGCCCCTCGCTCTCCCT
+
A<<FAA,,FFKA<FKKKKKKKAAFKFFKKAKKKFAF7AA,F<AKFFKAAAAKFKF,FFK,7,FFK,,<AA7FKFF<,7AFKKKKF7<((<F<(,A<AA<7
@ST-J00106:29:H3C5HBBXX:4:1101:3507:1000 1:N:0:TGACCA
AGTGAAAATAGCGTTAACTGATGACATTCCACAATTGTGATTTGTTTCTGCCCCACCCTAACTGATCAATGTACTTTGTAATCTCCCCCACCCTTAAGA
+
AAAFFKKKKKKKKKKKKKKKKKKKKKKKFKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKFKKK7FK<AAKKK,7FAKK,<FAFKKKKKKK<KKKFK
''')
    with open('data/S20_R2.fastq', 'w') as R2:
        R2.write('''\
@ST-J00106:29:H3C5HBBXX:4:1101:1172:1000 2:N:0:TGACCA
AGGATAAGCAGTCCAAGGAGGCATTGACCAAAACATGCAGTGAAATGTTGATGCGGAGCCTAAAGGACGGGGATCTGGGGGAGCCACGCCTGGAGGCG
+
A<,,,,,7A<F7,,,7,,,,,(,7,,,7,,,7<7,,7<7,,,,7,7A,,,,,,7,,,,(,AF,7,A,,,((,,77<<,7(,F7((A,(((7(<,,7A,
@ST-J00106:29:H3C5HBBXX:4:1101:1375:1000 2:N:0:TGACCA
TGAGAATCTATGGGCATGTGTAGTGTTACATAAGTGGGGTCACATACTGGATAGCAAAGTGCAAGACACAGACATCGTTCGGCATCCATGAGTAGAGTTG
+
<AA<FFFFKKKKKKFFKK7KKAKFF7FAFA,KAKKKAK<7AFKF,FAFFKKKKKKKKAFKKKFFKKKF,AFAF7AKFKA,,77AFA7FAFK<,FFAK<AF
@ST-J00106:29:H3C5HBBXX:4:1101:1395:1000 2:N:0:TGACCA
AGTCCTCTTGTCTTCTTCTGTTACCATGTTAATAGCCACACCCTTATGGCCAAACCATCCACCTCGACCAATTCTGTGGATACAGGTGGATGGTTTGGCC
+
AAFAFKKKFKKKKKKKKKKKKKKKKKKKFAKKKKKKKKKKK7F,AKKAKKKKKKKKK,FFKF<KAKKF<<AAF7FK7F<FAK<FFFAFF7<<KKAAFK7,
@ST-J00106:29:H3C5HBBXX:4:1101:1497:1000 2:N:0:TGACCA
AGGCGAATCGATGTGGCTCCCCCAGCCCCTTCCTGGGAGCATCCTAGAAAGACAGCGTGGAAACGCGCACGGCCTGGTGGTCCCGGTAGCGGCCATGGTG
+
AAFFFFAAFKKKKFKKKAKKKKKKKKKKKFKKK,7AAKFKK<KKFAKK,FAAKKAKKFK<FFKKKFF<(7<<F<FFAK7AAFKK<(<F,K7AFKF,AA<F
@ST-J00106:29:H3C5HBBXX:4:1101:1598:1000 2:N:0:TGACCA
AAAACTGAAAGTTCCAGGGGAAGTGGAAAGGATCACCACCTGCAGAGTCAGATATTCACTTCTGGTAACTCTGTGGACCTTTGGCAGGGGGCTTCAGA
+
AAAFFKKFKKKKKKKKKKAFFKKKKKKKFFKKKKKKKKAFKKKKAAFK7<KK7FFFAKKAFA,FKF7K,7FFFKFKKKAFF,AAKKKKFAAAFFAAFK
@ST-J00106:29:H3C5HBBXX:4:1101:1619:1000 2:N:0:TGACCA
TCCTCCTGCTTAAAACCCAAAAGGTCAGAAGGATCGTGAGGCCCCGCTTTCACGGTCTGTATTCGTACTGAAAATCAAGATCAAGCGAGCTTTTGCCCTT
+
,AAFFK<KKKFKKKKKKKKKKKAFFKKKAF7AA,AFKFFKKKF<<AFFKFKKKKKKK,,,FF,FFFKKFKKKKKAAFAK,,,7,,7AAKKA<FFF,FFKK
@ST-J00106:29:H3C5HBBXX:4:1101:1923:1000 2:N:0:TGACCA
TCACTTTCATGCCAGACTACAACTGTACCCCACCTCTTTGTGGGGTACAGTGGTAGTCTGGCATGAAAGTGAAGAACAAGGTAGGTTGGGGCTTCCTTGT
+
<AFFFKKKKKKKFKKKKFKKKKKKKKKFFFKKFF,A,A<<F7FA<,FKF,,,FFFKKKKAFKF7,FKK,7FFFAFKKKFAF7AFAF7FKKFFKKK<FKKK
@ST-J00106:29:H3C5HBBXX:4:1101:1964:1000 2:N:0:TGACCA
CCCTTAATATTTTTTCGTTCATTTCCATCTTGGTTAATCTGACAATTATGGGTCTTGGAGTTGCTCTTTGCAAGGAGTTTCTTTGTGGTATTCTCTATAT
+
<<FAFKKKKKAAKKKKKKKKKKKKFAFKAKKKKK7,FKAKK7<,AFF<KAKAKKKKK7AA<F,AFF7KKF,7A7<,<FKK<FKK7F7AAFFFAF<FKKKK
@ST-J00106:29:H3C5HBBXX:4:1101:2086:1000 2:N:0:TGACCA
TAGTTTTATATATGTGTATTTAATCCTGTCATTGTGTTCTTAGCTGGTTATTATGCTGTCTTGTTTGTTTGGCTGCTTTATAGTATCGCTGGTCTGCAT
+
,AFFFKKKKKKKKKKKKKF<KFKKKKKKKKFKKKKKKK7FKFKKKKK<F<<KKKKKAKKKKK7F7FA7,7FAFKK,FKKK7,FFKAAKF,,AFAKKKKA
@ST-J00106:29:H3C5HBBXX:4:1101:2187:1000 2:N:0:TGACCA
ATTGGATCGAAGGATCCAAAGTGTTTTTCCCGGTTGTGTCTGTGAGGGTGTTGCCAAAGGAGATTAACATTTATTTAGTGGACTGGGAAAGG
+
<AF<FK7FKKKKKKKKKKKKKKKKKKKFKFFFKKFAKKKKKKKFKKK<<KKKAKKKKKKKKKAF,FFK,F<AFAFKKFKFKFAFFAFFFKKK
@ST-J00106:29:H3C5HBBXX:4:1101:2207:1000 2:N:0:TGACCA
GAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATTAAAAAAAAAAAACAAGAGTAGTGTACGGAGCGTGGTTGTCTAGAGGTT
+
AAFFFKKKKKK7FKKKKKKKKKKKFKKKKKKKKKKKKKKKKKKKKKKFKKKKKKKKKKKKKKF7,,7,,7,,77,,7,,,,,((A7(,,77,,,,,,77,
@ST-J00106:29:H3C5HBBXX:4:1101:2289:1000 2:N:0:TGACCA
GAGGTACAAATACGCCCAGACTGCAAAGACGAAAAGAAACCCCGAACGCTTCAAAGCCAAGACACAGAGAAACATCTACAAGAATAAAGAAAATCCAGG
+
,AAF,FA,A7,,,,A<FFA,F,AFF,,<FA,F7FA,7FA,,AF<F7A,,777KK,7,,,7A,,A7FAFFF,,<F<,,AF,KA,7,A<FFAAFF,A<7,7
@ST-J00106:29:H3C5HBBXX:4:1101:2410:1000 2:N:0:TGACCA
TTGCAGTGACAGGTGCCCCGAGGAAAGCAGGCAAACATGCACTGGACAAAGACACATGCAGCCCTATTCTAAGAAGGACTTGACCTTTGAGTCTGAAAA
+
,A,FFFFKKK<FKKKKA<7<((<AFF<KAFFKKAFFFK,AFFFK7KKKFFAFKKKKK,FAFFFKFKK7A<FKKAFAFKFAF7F7<,A,AFFFF,FFFKK
@ST-J00106:29:H3C5HBBXX:4:1101:2512:1000 2:N:0:TGACCA
GCAGGTTAGAGAAGTGGGCCTCAGAAGTGCCAGGACATCATGAAGGGAGATAATCAGCTATGCCTCCCAGGACAAAGGGAGTGCCCTCTACTGAGTGGT
+
<AFFF77FFKKKFKFAAA<F(FKKKKKKKKKKK<KAF<<F7KKKFKKFKFKKK<7KKKAAFKAA,A(,,AFKKK77FFKKKKKFA7F<AAFFKA7AFFF
@ST-J00106:29:H3C5HBBXX:4:1101:2593:1000 2:N:0:TGACCA
AAGGTATCAACTATGGGATGGCCCGAGTCTTCTAGATCTGGCCTTGGCCAAACCCTTGGTCCTGATCTCCTGCCCCCCTCCTTCAAACACAAAGTTA
+
AAFFFKKKKKKFKKKFFFFKAFKAFKFFFKFFKKKKKKKKF<F<FAFFKKKKKAFAF,<<A7,7FFA,,,A<A<,<<,F<F7FAFKKKKKKKK7AFK
@ST-J00106:29:H3C5HBBXX:4:1101:2613:1000 2:N:0:TGACCA
GCTGGGATTACAGGCGTGAGCCACCGCGCCCGGCCTTGACATAATTTTCAAGAGCACTTTATGGCAGGGTTTGTAAATA
+
<AFFFKKKKKKKKKKKFFKKKKKKKKKKKFKKKKFFKKKKKKKKFAFKKFKKKKKKKAFFKF7FKKKFKKA<FKKKKKK
@ST-J00106:29:H3C5HBBXX:4:1101:2756:1000 2:N:0:TGACCA
CTTTGTGTTTGAGTGGGCTGAGGGCAGAAGTTATAGTCTCTTATAATCCAGTCTCAGAAGTAACTGAACAGCACTCCTAGCTCTTGTGGGTCACACATAC
+
AAFFFKKKKKKKKKKKKKAAKFKKKKKKKKF<FKAFKKKKKKK<FFKAFKK7AFAKKFKF7,FFAFAFAAKAFAA7FAKF7,AF<<AF,AAFKKAKKKKK
@ST-J00106:29:H3C5HBBXX:4:1101:2918:1000 2:N:0:TGACCA
TGCGCAGCAGCACTTTCTCTTCTTTTTCCTCCTGCCTCAAGAATAGCCCAGATGAGAAAAAAAGTGTTGTTCCTCATCCTGTTTCCATGTCCCTTTCTTT
+
AAFFFKFKKKKKKKKKKKKKKKKKKKKKKKKK<,AFKKKKKFKKKAFKKKKFKKKKKKK<FK,AFFFK,A<FFFFKKFKKKFK,FFFFKKKKKKKKKKKK
@ST-J00106:29:H3C5HBBXX:4:1101:3202:1000 2:N:0:TGACCA
GGCCCGTACCCCTCCGAGCAGACCTTTCCGCGTCCCAACATCGCCCCACCCCACTCCAGCCTGGGCGACAGAGCCTGACTCCATCGCAAAAA
+
,,,,,((,,7A,,,,(,,,,FK,,A,,,,,,(((,(F,,7F7FK,<<FA((<AA7FFF,AFAF7<FKKKKFKKFKAFKKFKK7FF,FKFFFK
@ST-J00106:29:H3C5HBBXX:4:1101:3344:1000 2:N:0:TGACCA
GGGGGGTGGTGAAGTATCTTTCTCTGTGGGTCTAGGGTCTCTCTGTCTCAGGGTCTGGGGTCTGCATCTTTAGGACCTCTGTCTCTCTCTGGTGTGTTTT
+
AAFAFAAAAAAKKKFKA,AA,A<FAF7AAK<FAKKFKFK<FAF7FKF7FKK<FFK7FKFFFF<FKKFA7<FFKKKK7,A,AFAF<FFAA7AKAFKAF7FF
@ST-J00106:29:H3C5HBBXX:4:1101:3385:1000 2:N:0:TGACCA
AAAACTGGAAGCATTCCCTTTGAAAACCGGCACAAAACAAGGATGCCCTCTCTCACCACTCCCATTCAACATAGTATTGGAAGTTCTAGCCAGGGCAATC
+
AAFFFFKFFKAFKF,,F<KFFFKFKKAAA<<A7KAFKKKK77F,7,,A7F7A,FKFKFA,AAFAFAAFKAA7,,FKKKF<FF,7,7AFK7,<7<,AAA7,
@ST-J00106:29:H3C5HBBXX:4:1101:3425:1000 2:N:0:TGACCA
TCAAAGTCGAGGTGCTAATGCAAAATTGATATCATACAGACTTAGAAATAAAGATGACATACTCAAACTATAGAAGGAGCCAAAAAGTGGTTCTGGT
+
,AFFFKKKKKKKKKKKKKKKKKKKKFKKFFFFKKKKKKAKKAKKKKKKFKKKKKKKKKKFFK<KKKKK,KKKKKKKKKKKKKKKKKKKKKAKKFF<A
@ST-J00106:29:H3C5HBBXX:4:1101:3446:1000 2:N:0:TGACCA
GTCTAAGTGGCGATCCTAGTCCTTTACCTGGGACTGTCTTGCCTTTTTAGCCCCCAGGGTGTTGAAGGTAAACGTCTAGGGT
+
A,A<FAA,,<7(,,7A,A,,,,,,AF,,,7,(,,,7,7<,7,,,,,,,,A<AK((,(((77,,,,,<<A77,,,77,,7<7,
@ST-J00106:29:H3C5HBBXX:4:1101:3466:1000 2:N:0:TGACCA
CTCCGCACATGTGAGGGAGAGCGAGGGGCCACTGGGATTGTGTGGAGGTGAGATCAGAGGTCGAACAGTTGGGAGGAGCAGTGCTGTTCACCTGTAGGTC
+
,,A<FFKF7,A7AFKKKKK,AAKKKFF<F,7,7F,F77,,AF,F77FFAA,AFKKAFAFF,<AF,FA7F,FA7<7AFFKKK7FKA,,,<AFAKKFAA7AA
@ST-J00106:29:H3C5HBBXX:4:1101:3507:1000 2:N:0:TGACCA
AGGGTGGGGAGAATTACAAAGAACCTTCTTAAGAGTGGGGGAGATTACAAAGAACATTCTTAAGGGTGAGGGAGATTACAAAGTACATTGATCAGTTAGG
+
<A<FAF7A(<AFFKKA<KKKAFKF7<FKFKKKKK7,7,,(AFAKFKKFKAKKAKAFK7A,,FFA,<F7,,AAFFFK<FFAKKAFKF<,AAAFAFAK7FKK
''')


#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil
import subprocess
import pytest
from sos import execute_workflow


@pytest.mark.skipif(
    not shutil.which("Rscript") or "TRAVIS" in os.environ, reason="R not installed"
)
def test_r_library():
    """Test target R_Library"""
    execute_workflow(
        """
        [default]
        depends: R_library("dplyr", autoinstall=True)
        R:
            library('dplyr')
        """
    )


@pytest.mark.skipif(not shutil.which("Rscript"), reason="R not installed")
def test_depends_r_library():
    """Testing depending on R_library"""
    # first remove xtable package
    subprocess.call("R CMD REMOVE xtable", shell=True)
    execute_workflow(
        """
        [0]

        depends: R_library('xtable', autoinstall=True)
        R:
        library('xtable')
        ## Demonstrate data.frame
        tli.table <- xtable(cars)
        """
    )


@pytest.mark.skipif(not shutil.which("Rscript"), reason="R not installed")
def test_reexecution():
    """Test re-execution of steps with R_library"""
    subprocess.call("R CMD REMOVE xtable", shell=True)
    wf = """
    [1]
    depends: R_library("xtable", autoinstall=True)
    output: '1.txt'
    run: expand=True
        sleep 5
        touch {_output}
    """
    execute_workflow(wf)
    execute_workflow(wf)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import ast

from collections.abc import Mapping, Sequence
from typing import Any, Dict, Optional

from .eval import SoS_eval, accessed_vars, used_in_func, SoS_exec
from .parser import SoS_Step
from .targets import dynamic, remote, sos_targets, sos_step, named_output, file_target
from .utils import env
from .executor_utils import __null_func__, prepare_env, strip_param_defs
from .syntax import SOS_TARGETS_OPTIONS


def get_param_of_function(name, param_list, extra_dict={}):
    tree = ast.parse(f'__null_func__({param_list})')
    # x.func can be an attribute (e.g. a.b()) and do not have id
    funcs = [
        x for x in ast.walk(tree) if x.__class__.__name__ == 'Call' and
        hasattr(x.func, 'id') and x.func.id == name
    ]
    params = []
    for func in funcs:
        for arg in func.args:
            try:
                params.append([ast.literal_eval(arg)])
            except Exception as e:
                if 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                        'SOS_DEBUG']:
                    env.log_to_file(
                        'STEP',
                        f'Failed to evaluate parameter of function {name} from {param_list}: {e}'
                    )
                try:
                    params.append([
                        eval(
                            compile(
                                ast.Expression(body=arg),
                                filename='<string>',
                                mode="eval"), extra_dict)
                    ])
                except Exception as e:
                    if 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                            'SOS_DEBUG']:
                        env.log_to_file(
                            'STEP',
                            f'Failed to evaluate parameter of function {name} from {param_list}: {e}'
                        )
        for kwarg in func.keywords:
            try:
                params.append([kwarg.arg, ast.literal_eval(kwarg.value)])
            except Exception as e:
                if 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                        'SOS_DEBUG']:
                    env.log_to_file(
                        'STEP',
                        f'Failed to evaluate parameter of function {name} from {param_list}: {e}'
                    )
                try:
                    params.append([
                        kwarg.arg,
                        eval(
                            compile(
                                ast.Expression(body=kwarg.value),
                                filename='<string>',
                                mode="eval"), extra_dict)
                    ])
                except Exception as e:
                    if 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                            'SOS_DEBUG']:
                        env.log_to_file(
                            'STEP',
                            f'Failed to evaluate parameter of function {name} from {param_list}: {e}'
                        )
    return params


def get_names_of_param(name, param_list, extra_dict={}):
    tree = ast.parse(f'__null_func__({param_list})')
    kwargs = [
        x for x in ast.walk(tree)
        if x.__class__.__name__ == 'keyword' and x.arg == name
    ]

    values = []
    for kwarg in kwargs:
        # Python 3.8 class name is named Constant, not Str
        values.extend([
            x.s
            for x in ast.walk(kwarg.value)
            if x.__class__.__name__ in ('Str',
                                        'Constant') and isinstance(x.s, str)
        ])
    return values


def get_num_of_args_and_names_of_kwargs(param_list):
    tree = ast.parse(f'__null_func__({param_list})')
    # func is the __null_func__
    func = tree.body[0].value
    nargs = len(func.args)
    keywords = [x.arg for x in func.keywords]
    return nargs, keywords


def find_statement(section, name):
    stmt_idx = [
        idx for idx, x in enumerate(section.statements)
        if x[0] == ':' and x[1] == name
    ]
    if not stmt_idx:
        return None
    elif len(stmt_idx) == 1:
        return stmt_idx[0]
    else:
        raise RuntimeError(
            f'More than one step {name} statement are specified in step {section.step_name()}'
        )


def no_output_from(*args, **kwargs):
    raise SyntaxError(
        'Function output_from can only be used in input or depends statements')


def no_named_output(*args, **kwargs):
    raise SyntaxError(
        'Function named_output can only be used in input or depends statements')


def no_sos_step(*args, **kwargs):
    raise SyntaxError('Target sos_step can only be used in depends statements')


def no_sos_variable(*args, **kwargs):
    raise SyntaxError(
        'Target sos_variable can only be used in depends statements')


def get_changed_vars(section: SoS_Step):
    '''changed vars are variables that are "shared" and therefore "provides"
    to others '''
    if 'shared' not in section.options:
        return set()

    changed_vars = set()
    svars = section.options['shared']
    if isinstance(svars, str):
        changed_vars.add(svars)
        svars = {svars: svars}
    elif isinstance(svars, Sequence):
        for item in svars:
            if isinstance(item, str):
                changed_vars.add(item)
            elif isinstance(item, Mapping):
                changed_vars |= set(item.keys())
            else:
                raise ValueError(
                    f'Option shared should be a string, a mapping of expression, or list of string or mappings. {svars} provided'
                )
    elif isinstance(svars, Mapping):
        changed_vars |= set(svars.keys())
    else:
        raise ValueError(
            f'Option shared should be a string, a mapping of expression, or list of string or mappings. {svars} provided'
        )
    return changed_vars


def get_environ_vars(section):
    # environ variables are variables should be inserted from outside
    # which is basically the sos_variable(var)
    environ_vars = set()
    depends_idx = find_statement(section, 'depends')
    if depends_idx is not None:
        value = section.statements[depends_idx][2]
        args = get_param_of_function(
            'sos_variable', value, extra_dict=env.sos_dict.dict())
        for arg in args:
            if len(arg) == 2:
                raise SyntaxError(
                    'sos_variable does not accept keyword argument')
            environ_vars.add(arg[0])
    if section.task:
        # 1281
        environ_vars |= accessed_vars(section.task_params, mode='eval')
    return environ_vars


def get_all_used_vars(section):
    '''Get variables which are variables used by input statement and statements before it'''
    all_used_vars = set()
    for statement in section.statements:
        if statement[0] == '=':
            all_used_vars |= accessed_vars('='.join(statement[1:3]))
        elif statement[0] == '!':
            all_used_vars |= accessed_vars(statement[1])
        elif statement[0] == ':':
            all_used_vars |= accessed_vars(statement[2], mode='eval')
            if statement[1] != 'input':
                continue
            if 'paired_with' in statement[2]:
                try:
                    pws = get_names_of_param(
                        'paired_with',
                        statement[2],
                        extra_dict=env.sos_dict.dict())
                    all_used_vars |= set(pws)
                except Exception as e:
                    raise ValueError(
                        f'Failed to parse parameter paired_with: {e}')
            if 'group_with' in statement[2]:
                try:
                    pws = get_names_of_param(
                        'group_with',
                        statement[2],
                        extra_dict=env.sos_dict.dict())
                    all_used_vars |= set(pws)
                except Exception as e:
                    raise ValueError(
                        f'Failed to parse parameter group_with: {e}')
            if 'for_each' in statement[2]:
                try:
                    pws = get_names_of_param(
                        'for_each',
                        statement[2],
                        extra_dict=env.sos_dict.dict())
                    for pw in pws:
                        all_used_vars |= set(pw.split(','))
                except Exception as e:
                    raise ValueError(f'Failed to parse parameter for_each: {e}')
    if section.task:
        all_used_vars |= accessed_vars(section.task)

    # now we have a list of global variables that are actually used in the functions
    # this is specifically designed to handle the last case in #1225
    func_with_vars = [
        y for x, y in used_in_func(section.global_stmts).items()
        if x in all_used_vars
    ]
    return set.union(all_used_vars, *func_with_vars)


def get_signature_vars(section):
    '''Get signature variables which are variables that will be
    saved with step signatures'''

    # signature vars should contain parameters defined in global section
    # #1155, we however need to keep local parameters
    signature_vars = set(section.parameters.keys()) - set(
        section.global_parameters.keys())
    signature_vars |= set(
        section.global_parameters.keys()
        & accessed_vars(strip_param_defs(section.global_stmts)))

    input_idx = find_statement(section, 'input')
    after_input_idx = 0 if input_idx is None else input_idx + 1

    for statement in section.statements[after_input_idx:]:
        if statement[0] == '=':
            signature_vars |= accessed_vars('='.join(statement[1:3]))
        elif statement[0] == '!':
            signature_vars |= accessed_vars(statement[1])
    # finally, tasks..
    if section.task:
        signature_vars |= accessed_vars(section.task)

    return {x for x in signature_vars if not x.startswith('__')}


def get_step_depends(section):

    step_depends: sos_targets = sos_targets()
    dynamic_depends = True

    input_idx = find_statement(section, 'input')
    depends_idx = find_statement(section, 'depends')
    for stmt_idx in ([] if input_idx is None else [input_idx]) + (
        [] if depends_idx is None else [depends_idx]):
        # input statement
        stmt = section.statements[stmt_idx][2]
        if 'sos_step' in stmt:
            step_depends.extend([sos_step(x) for x in get_sos_step_steps(stmt)])
        if 'output_from' in stmt:
            step_depends.extend([
                sos_step(x)
                for x in get_output_from_steps(stmt, section.last_step)
            ])
        if 'named_output' in stmt:
            # there can be multiple named_output calls
            pars = get_param_of_function(
                'named_output', stmt, extra_dict=env.sos_dict.dict())
            for par in pars:
                # a single argument
                if len(par) == 1:
                    if not isinstance(par[0], str):
                        raise ValueError(
                            f'Value for named_output can only be a name (str): {par[0]} provided'
                        )
                    step_depends.extend(named_output(par[0]))
                else:
                    if par[0] in SOS_TARGETS_OPTIONS:
                        continue
                    elif par[0] == 'name':
                        if not isinstance(par[1], str):
                            raise ValueError(
                                f'Value for named_output can only be a name (str): {par[1]} provided'
                            )
                        step_depends.extend(named_output(par[1]))
                    else:
                        raise ValueError(
                            f'Unacceptable keyword argument {par[0]} for named_output()'
                        )

    if depends_idx is not None:
        value = section.statements[depends_idx][2]
        try:
            svars = ['output_from', 'named_output']
            old_values = {
                x: env.sos_dict.dict()[x]
                for x in svars
                if x in env.sos_dict.dict()
            }
            # output_from and named_output has been processed
            env.sos_dict.quick_update({
                'output_from': lambda *args, **kwargs: None,
                'named_output': lambda *args, **kwargs: None,
                'traced': lambda *args, **kwargs: sos_targets(*args, **kwargs)
            })
            args, kwargs = SoS_eval(
                f'__null_func__({value})', extra_dict=env.sos_dict.dict())
            if any(isinstance(x, (dynamic, remote)) for x in args):
                dynamic_depends = True
            else:
                step_depends.extend(sos_targets(*args, **kwargs))
                dynamic_depends = False
        except SyntaxError:
            raise
        except Exception as e:
            if 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                    'SOS_DEBUG']:
                env.log_to_file(
                    'STEP',
                    f"Args {value} in depends cannot be determined: {e}")
        finally:
            [env.sos_dict.dict().pop(x) for x in svars]
            env.sos_dict.quick_update(old_values)
    return step_depends, dynamic_depends


def get_step_input(section, default_input):
    '''Find step input
    '''
    step_input: sos_targets = sos_targets()
    dynamic_input = True

    # look for input statement.
    input_idx = find_statement(section, 'input')
    # #1270. If there is any statement before input:, it might create input or remove input,
    # which essentially make input undetermined before actually run it.
    if input_idx is None or (input_idx != 0 and any(
            x[0] == '!' for x in section.statements[:input_idx])):
        return step_input, dynamic_input

    # input statement
    stmt = section.statements[input_idx][2]
    try:
        svars = ['output_from', 'named_output', 'sos_step', 'sos_variable']
        old_values = {
            x: env.sos_dict.dict()[x]
            for x in svars
            if x in env.sos_dict.dict()
        }
        env.sos_dict.quick_update({
            'output_from': lambda *args, **kwargs: None,
            'named_output': lambda *args, **kwargs: None,
            'traced': lambda *args, **kwargs: sos_targets(*args, **kwargs),
            'sos_step': no_sos_step,
            'sos_variable': no_sos_variable,
        })
        args, kwargs = SoS_eval(
            f'__null_func__({stmt})', extra_dict=env.sos_dict.dict())
        if not args:
            if default_input is None:
                step_input = sos_targets()
            else:
                step_input = default_input
        elif not any(isinstance(x, (dynamic, remote)) for x in args):
            step_input = sos_targets(*args)
    except SyntaxError:
        raise
    except Exception as e:
        # if anything is not evalutable, keep Undetermined
        env.logger.debug(
            f'Input of step {section.name if section.index is None else f"{section.name}_{section.index}"} is set to Undertermined: {e}'
        )
        # expression ...
        step_input = sos_targets(_undetermined=stmt)
    finally:
        [env.sos_dict.dict().pop(x) for x in svars]
        env.sos_dict.quick_update(old_values)
    return step_input, dynamic_input


def get_step_output(section, default_output, analysis_type):
    '''determine step output'''
    #
    # There are three analysis_style:
    #
    #  'default': used for initial pass, we do not have any
    #      default_output, and we just try to get some information
    #      if we can.
    #
    #  'forward': used when adding forward step. We do not care
    #       about default_output.
    #
    #  'backward': used when adding as backwwad step. Smetimes we have
    #       to figure out output.
    #
    step_output: sos_targets = sos_targets()

    if 'provides' in section.options and default_output:
        step_output = default_output

    # look for output statement.
    output_idx = find_statement(section, 'output')
    if output_idx is None:
        return step_output

    # if the step is referred to by named_output, sos_step etc, we do not
    # care about their output. In terms of bug #1379, perhaps the step would
    # be added to the DAG multiple times, it will be run only once.
    if analysis_type == 'backward' and (default_output is None or all(
            isinstance(x, sos_step) for x in default_output)):
        return step_output

    # # if the output is of type named_output, we do not need to care about
    # # the exact output either. We just need to check if there are other
    # # named output
    n_args, name_kwargs = get_num_of_args_and_names_of_kwargs(
        section.statements[output_idx][2])
    name_kwargs = [x for x in name_kwargs if x not in SOS_TARGETS_OPTIONS]
    if name_kwargs:
        step_output.extend([named_output(x) for x in name_kwargs])

    # now, if we are referred by a filename, we have to figure out what
    # these filenames are... if there is only one argument, let us
    # be lazy and assume that the output is the one we want.. Note that
    # we are assuming that the step does not have any substep.
    # if n_args + len(name_kwargs) == 1:
    #     return step_output

    # if we do have output, we have to evaluate starting from input...
    for statement in section.statements[:output_idx + 1]:
        if statement[0] == ':' and statement[1] == 'depends':
            continue
        if statement[0] == '!':
            if analysis_type == 'backward':
                try:
                    SoS_exec(statement[1], return_result=False)
                except Exception as e:
                    raise f'Failed to evaluate an statement "{value}" of an auxiliary step: {e}'
            continue

        if statement[1] == 'input' and analysis_type != 'backward':
            continue

        value = statement[2]
        try:
            svars = ['output_from', 'named_output', 'sos_step', 'sos_variable']
            old_values = {
                x: env.sos_dict.dict()[x]
                for x in svars
                if x in env.sos_dict.dict()
            }
            env.sos_dict.quick_update({
                'output_from': no_output_from,
                'named_output': no_named_output,
                'sos_step': no_sos_step,
                'sos_variable': no_sos_variable,
            })
            args, kwargs = SoS_eval(
                f'__null_func__({value})', extra_dict=env.sos_dict.dict())
            if any(isinstance(x, (dynamic, remote)) for x in args):
                raise ValueError(
                    f'Auxiliary step does not allow dynamic or remote input or output: {value} provided'
                )
            if statement[1] == 'input':
                step_input = sos_targets(
                    *args, **{
                        x: y
                        for x, y in kwargs.items()
                        if x not in SOS_TARGETS_OPTIONS
                    })
                env.sos_dict.set('_input', step_input)
                env.sos_dict.set('step_input', step_input)
            else:
                step_output = sos_targets(
                    *args, **{
                        x: y
                        for x, y in kwargs.items()
                        if x not in SOS_TARGETS_OPTIONS
                    })
            if kwargs:
                step_output.extend(named_output(x) for x in kwargs.keys())
        except SyntaxError:
            raise
        except Exception as e:
            if 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                    'SOS_DEBUG']:
                env.log_to_file('STEP',
                                f"Args {value} cannot be determined: {e}")
            # usually we want to get the exact output. However in the case when
            # the step is referred by named_output(), it is ok for us to not
            # know the details.
            if analysis_type == 'backward' and (n_args > 0 or not all(
                    isinstance(x, named_output) for x in default_output)):
                raise RuntimeError(
                    f'Failed to determine input "{value}" of an auxiliary step: {e}'
                )
        finally:
            [env.sos_dict.dict().pop(x) for x in svars]
            env.sos_dict.quick_update(old_values)

    if 'provides' in section.options and default_output is not None:
        for out in default_output:
            # 981
            if not isinstance(out, sos_step) and out not in step_output:
                raise ValueError(
                    f'Defined output fail to produce expected output: {step_output} generated, {default_output} expected.'
                )
    return step_output


def get_sos_step_steps(stmt):
    '''
    Extract sos_step(x) from statement
    '''
    opt_values = get_param_of_function(
        'sos_step', stmt, extra_dict=env.sos_dict.dict())
    for value in opt_values:
        if len(value) != 1:
            raise ValueError('sos_step only accept one and only one parameter')
    return [x[0] for x in opt_values]


def get_output_from_steps(stmt, last_step):
    '''
    Extract output_from(1), output_from('step_1'), and output_from([1, 2])
    to determine dependent steps
    '''
    opt_values = get_param_of_function(
        'output_from', stmt, extra_dict=env.sos_dict.dict())

    def step_name(val):
        if isinstance(val, str):
            return val
        elif isinstance(val, int):
            if val == -1:
                if last_step is None:
                    # there is a case where a regular step is checked as auxiliary step.
                    # we will postpone the decision later because the step might not be
                    # used as such
                    return None
                return last_step
            if '_' in env.sos_dict['step_name']:
                return f"{env.sos_dict['step_name'].rsplit('_',1)[0]}_{val}"
            else:
                return str(val)
        else:
            raise ValueError(f'Invalid value {val} for output_from() function')

    res = []
    for value in opt_values:
        if len(value) == 1:
            # regular argument
            value = value[0]
        elif value[0] == 'steps':
            value = value[1]
        elif value[0] in SOS_TARGETS_OPTIONS:
            continue
        else:
            raise ValueError(
                f'Unacceptable keyword argument {value[0]} for function output_from'
            )
        if isinstance(value, (int, str)):
            res.append(step_name(value))
        elif isinstance(value, Sequence):
            res.extend([step_name(x) for x in value])
        else:
            raise ValueError(f'Invalid value for input option from {value}')
    return [x for x in res if x is not None]


# analysis_cache = {}


def analyze_section(section: SoS_Step,
                    default_input: Optional[sos_targets] = None,
                    default_output: Optional[sos_targets] = None,
                    context=None,
                    analysis_type='initial') -> Dict[str, Any]:
    '''Analyze a section for how it uses input and output, what variables
    it uses, and input, output, etc.'''

    # analysis_key = (section.md5, section.step_name(),
    #     default_input.target_name() if hasattr(default_input, 'target_name') else '',
    #     default_output.target_name() if hasattr(default_output, 'target_name') else '', vars_and_output_only)
    #if analysis_key in analysis_cache:
    #    return analysis_cache[analysis_key]

    # use a fresh env for analysis
    new_env, old_env = env.request_new()
    try:
        prepare_env(section.global_def, section.global_vars,
                    {} if context is None else context)

        env.sos_dict.set('step_name', section.step_name())
        env.sos_dict.set('__null_func__', __null_func__)
        if 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
            env.log_to_file(
                'STEP',
                f'Analyzing {section.step_name()} {"(output only)" if analysis_type == "initial" else ""}'
            )

        res = {
            'step_name':
                section.step_name(),
            'step_output':
                get_step_output(section, default_output, analysis_type),
            # variables starting with __ are internals...
            'environ_vars':
                get_environ_vars(section),
            'signature_vars':
                get_signature_vars(section),
            'changed_vars':
                get_changed_vars(section)
        }
        if analysis_type != 'initial':
            inps = get_step_input(section, default_input)
            res['step_input'] = inps[0]
            res['dynamic_input'] = inps[1]
            deps = get_step_depends(section)
            res['step_depends'] = deps[0]
            res['dynamic_depends'] = deps[1]
    # analysis_cache[analysis_key] = res
    finally:
        # restore env
        env.restore_to_old(new_env, old_env)

    # #1225
    # The global section can contain a lot of variables, some of which can be large. Here we
    # found all variables that will be used in the step, including ones used in substep (signature_vars)
    # and ones that will be used in input statement etc.
    section.global_vars = {
        x: y
        for x, y in section.global_vars.items()
        if x in get_all_used_vars(section)
    }
    return res



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import unittest

from sos.parser import SoS_Script
from sos.targets import file_target
from sos.utils import env
from sos.workflow_executor import Base_Executor


class TestPython3Target(unittest.TestCase):

    def setUp(self):
        env.reset()
        self.temp_files = []

    def tearDown(self):
        for f in self.temp_files:
            if file_target(f).exists():
                file_target(f).unlink()

    def test_py_module(self):
        '''Test target Py_Module'''
        if file_target('report.md').exists():
            file_target('report.md').unlink()
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate', autoinstall=True)
from tabulate import tabulate
table = [["Sun",696000,1989100000],["Earth",6371,5973.6],
    ["Moon",1737,73.5],["Mars",3390,641.85]]
report: output='report.md', expand=True
    {tabulate(table)}
''')
        wf = script.workflow()
        Base_Executor(wf).run()
        self.assertTrue(file_target('report.md').target_exists('target'))
        with open('report.md') as rep:
            self.assertEqual(
                rep.read().strip(), '''
-----  ------  -------------
Sun    696000     1.9891e+09
Earth    6371  5973.6
Moon     1737    73.5
Mars     3390   641.85
-----  ------  -------------
'''.strip())

    def test_py_module_with_version(self):
        '''Test target Py_Module'''
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate', version='2.0', autoinstall=True)
''')
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)
        #
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate>=2.0')
''')
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)
        #
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate==20.0')
''')
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)
        #
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate<2.0')
''')
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_upgrade_py_module(self):
        '''Test upgrade py module #1246'''
        # first install tabulate == 0.7.5
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate==0.7.5', autoinstall=True)
''')
        wf = script.workflow()
        Base_Executor(wf).run()
        # test should pass
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate'), Py_Module('tabulate==0.7.5')
''')
        wf = script.workflow()
        Base_Executor(wf).run()
        # test for newer version should fail
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate==0.8.3')
''')
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)
        # auto install should work
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate==0.8.3', autoinstall=True)
''')
        wf = script.workflow()
        Base_Executor(wf).run()
        # test for old version should fail
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate'), Py_Module('tabulate==0.8.3')
''')
        wf = script.workflow()
        Base_Executor(wf).run()
        # test for old version should fail
        script = SoS_Script(r'''
[10]
depends: Py_Module('tabulate==0.7.5')
''')
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import shutil

import pytest

from sos import execute_workflow


@pytest.mark.skipif(not shutil.which('python2.7'), reason = 'Skip test because of no python2.7 installation')
def test_python2():
    execute_workflow(r'''
        [0]
        python2: expand='${ }'
        a = {'1', '2'}
        print a
        ''')



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import glob
import os
import subprocess
import sys
import textwrap
import time

import pytest

from sos import execute_workflow
from sos._version import __version__
from sos.parser import SoS_Script
from sos.targets import file_target, sos_targets
from sos.utils import env
# if the test is imported under sos/test, test interacive executor
from sos.workflow_executor import Base_Executor


def multi_attempts(fn):

    def wrapper(*args, **kwargs):
        for n in range(4):
            try:
                fn(*args, **kwargs)
                break
            except Exception:
                if n > 1:
                    raise

    return wrapper


def test_command_line():
    '''Test command line arguments'''
    with open('test_cl.sos', 'w') as cl:
        cl.write('''\
#!/usr/bin/env sos-runner
#fileformat=SOS1.0

[L]
a =1
''')
    result = subprocess.check_output(
        'sos --version', stderr=subprocess.STDOUT, shell=True).decode()
    assert result.startswith('sos {}'.format(__version__))
    assert subprocess.call(
        'sos', stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos -h',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos run -h',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    #
    assert subprocess.call(
        'sos run test_cl -w -W',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 1
    assert subprocess.call(
        'sos-runner -h',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos dryrun -h',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos dryrun test_cl',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos dryrun test_cl.sos',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos dryrun test_cl L',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos-runner test_cl L',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    # script help
    assert subprocess.call(
        'sos-runner test_cl -h',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos convert -h',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0


def test_func_def(temp_factory):
    '''Test defintion of function that can be used by other steps'''
    temp_factory('aa.txt', 'ab.txt')
    execute_workflow(
        r"""
        def myfunc(a):
            sum(range(5))
            return ['a' + x for x in a]

        [0: shared={'test':'step_input'}]
        input: myfunc(['a.txt', 'b.txt'])
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['test'] == ['aa.txt', 'ab.txt']


def test_input(temp_factory):
    '''Test input specification'''
    temp_factory('test_input.txt', 'test_input1.txt')
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        input: '*.txt'
        output: [x + '.res' for x in _input]
        """,
        options={'run_mode': 'dryrun'})
    assert file_target('test_input.txt.res').resolve() in env.sos_dict['res']
    assert file_target('test_input1.txt.res').resolve() in env.sos_dict['res']


def test_for_each(temp_factory):
    '''Test for_each option of input'''
    temp_factory('a.txt', 'b.txt', 'a.pdf')
    execute_workflow(r"""
        [0: shared=['counter', 'all_names', 'all_loop']]
        files = ['a.txt', 'b.txt']
        names = ['a', 'b', 'c']
        c = ['1', '2']
        counter = 0
        all_names = ''
        all_loop = ''

        input: 'a.pdf', files, group_by='single', paired_with='names', for_each='c'

        all_names += str(_names[0]) + " "
        all_loop += str(_c) + " "

        counter = counter + 1
        """)
    assert env.sos_dict['counter'] == 6
    assert env.sos_dict['all_names'] == "a b c a b c "
    assert env.sos_dict['all_loop'] == "1 1 1 2 2 2 "

def test_for_each_same_level(temp_factory):
    '''Test for_each option of input'''
    temp_factory('a.txt', 'b.txt', 'a.pdf')
    # test same-level for loop and parameter with nested list
    execute_workflow(
        r"""
        [0: shared=['processed']]
        files = ['a.txt', 'b.txt']
        par = [(1, 2), (1, 3), (2, 3)]
        res = ['p1.txt', 'p2.txt', 'p3.txt']
        processed = []

        input: files, for_each='par,res'
        output: res, group_by=1

        processed.append((_par, _res))
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['processed'] == [((1, 2), 'p1.txt'), ((1, 3), 'p2.txt'),
                                         ((2, 3), 'p3.txt')]

def test_for_each_dataframe(temp_factory):
    '''Test for_each option of input'''
    temp_factory('a.txt', 'b.txt', 'a.pdf')
    # test for each for pandas dataframe
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        data = pd.DataFrame([(1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])

        input: for_each='data'
        output: f"{_data['A']}_{_data['B']}_{_data['C']}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['1_2_Hello.txt', '2_4_World.txt']

def test_for_each_dict(temp_factory):
    '''Test for_each option of input'''
    # test dictionary format of for_each
    temp_factory('a.txt', 'b.txt', 'a.pdf')
    execute_workflow(r"""
        [0: shared=['counter', 'all_names', 'all_loop']]
        files = ['a.txt', 'b.txt']
        names = ['a', 'b', 'c']
        counter = 0
        all_names = ''
        all_loop = ''

        input: 'a.pdf', files, group_by='single', paired_with='names', for_each={'c':  ['1', '2']}

        all_names += str(_names[0]) + " "
        all_loop += c + " "

        counter = counter + 1
        """)
    assert env.sos_dict['counter'] == 6
    assert env.sos_dict['all_names'] == "a b c a b c "
    assert env.sos_dict['all_loop'] == "1 1 1 2 2 2 "


def test_for_each_multikey_dict(temp_factory):
    '''Test for_each option of input'''
    # test multi-key dictionary format of for_each
    temp_factory('a.txt')
    execute_workflow(r"""
        import itertools
        [0: shared=['counter', 'all_names', 'all_loop']]
        parameter: n = [300, 100]
        parameter: p = [50, 200, 100]
        parameter: outfile = ['1', '2', '3', '4', '5', '6']
        counter = 0
        all_names = ''
        all_loop = ''
        input: 'a.txt', group_by='single', for_each={'_n,_p': [(_n,_p) for _n,_p in itertools.product(n,p) if _n > _p]}

        all_names += outfile[_index] + " "
        all_loop += '{} {} '.format(_n, _p)
        counter = counter + 1
        """)
    assert env.sos_dict['counter'] == 4
    assert env.sos_dict['all_names'] == "1 2 3 4 "
    assert env.sos_dict['all_loop'] == "300 50 300 200 300 100 100 50 "


def test_for_each_nested_list(temp_factory):
    '''Test for_each option of input'''
    # test dictionary format of for_each
    temp_factory('p1.txt', 'p2.txt', 'p3.txt', 'a.txt', 'b.txt')
    # test same-level for loop and parameter with nested list
    execute_workflow(
        r"""
        [0: shared=['processed']]
        files = ['a.txt', 'b.txt']
        processed = []

        input: files, for_each={'par':[(1, 2), (1, 3), (2, 3)], 'res': ['p1.txt', 'p2.txt', 'p3.txt']}
        output: res

        processed.append((par, res))
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['processed'] == [((1, 2), 'p1.txt'), ((1, 3), 'p2.txt'),
                                         ((2, 3), 'p3.txt')]

def test_for_each_pandas(temp_factory):
    '''Test for_each option of input'''
    # test for each for pandas dataframe
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        input: for_each={'data': pd.DataFrame([(1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])}
        output: f"{data['A']}_{data['B']}_{data['C']}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['1_2_Hello.txt', '2_4_World.txt']

def test_for_each_series():
    '''Test for_each option of input'''
    # support for pands Series and Index types
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        data = pd.DataFrame([(1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])
        input: for_each={'A': data['A']}
        output: f"a_{A}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['a_1.txt', 'a_2.txt']

def test_for_each_dataframe_row():
    '''Test for_each option of input'''
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        data = pd.DataFrame([(1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])
        data.set_index('C', inplace=True)
        input: for_each={'A': data.index}
        output: f"{A}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['Hello.txt', 'World.txt']

def test_for_each_each_series():
    # test for each of Series
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        data = pd.DataFrame([(0, 1, 'Ha'), (1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])

        data.set_index('A', inplace=True)
        data = data.tail(2)
        input: for_each={'A': data['B']}
        output: f"{A}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['2.txt', '4.txt']

def test_for_each_iterable():
    # test iterable
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        data = pd.DataFrame([(0, 1, 'Ha'), (1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])

        data.set_index('A', inplace=True)
        data = data.tail(2)
        input: for_each={'A,B': zip(data['B'],data['C'])}
        output: f"{A}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['2.txt', '4.txt']


def test_for_each_as_target_property(temp_factory):
    '''Test for_each option of input'''
    temp_factory('a.txt', 'b.txt', 'a.pdf')
    execute_workflow(r"""
        [0: shared=['counter', 'all_names', 'all_loop']]
        files = ['a.txt', 'b.txt']
        names = ['a', 'b', 'c']
        c = ['1', '2']
        counter = 0
        all_names = ''
        all_loop = ''

        input: 'a.pdf', files, group_by='single', paired_with='names', for_each='c'

        all_names += str(_input._names) + " "
        all_loop += str(_input._c) + " "

        counter = counter + 1
        """)
    assert env.sos_dict['counter'] == 6
    assert env.sos_dict['all_names'] == "a b c a b c "
    assert env.sos_dict['all_loop'] == "1 1 1 2 2 2 "


def test_for_each_as_target_property_same_level_loop(temp_factory):
    '''Test for_each option of input'''
    # test same-level for loop and parameter with nested list
    temp_factory('a.txt', 'b.txt', 'p1.txt', 'p2.txt', 'p3.txt')
    execute_workflow(
        r"""
        [0: shared=['processed']]
        files = ['a.txt', 'b.txt']
        par = [(1, 2), (1, 3), (2, 3)]
        res = ['p1.txt', 'p2.txt', 'p3.txt']
        processed = []

        input: files, for_each='par,res'
        output: res, group_by=1

        print([x._dict for x in step_input._groups])
        processed.append((_input._par, _input._res))
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['processed'] == [((1, 2), 'p1.txt'), ((1, 3), 'p2.txt'),
                                         ((2, 3), 'p3.txt')]


def test_for_each_as_target_property_pandas():
    # test for each for pandas dataframe
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        data = pd.DataFrame([(1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])

        input: for_each='data'
        print([x._dict for x in step_input._groups])
        output: f"{_input._data['A']}_{_input._data['B']}_{_input._data['C']}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['1_2_Hello.txt', '2_4_World.txt']

def test_for_each_as_target_property_dict(temp_factory):
    '''Test for_each option of input'''
    # test dictionary format of for_each
    temp_factory('a.txt', 'b.txt', 'a.pdf')
    execute_workflow(r"""
        [0: shared=['counter', 'all_names', 'all_loop']]
        files = ['a.txt', 'b.txt']
        names = ['a', 'b', 'c']
        counter = 0
        all_names = ''
        all_loop = ''

        input: 'a.pdf', files, group_by='single', paired_with='names', for_each={'c':  ['1', '2']}

        all_names += str(_input._names) + " "
        all_loop += _input.c + " "

        counter = counter + 1
        """)
    assert env.sos_dict['counter'] == 6
    assert env.sos_dict['all_names'] == "a b c a b c "
    assert env.sos_dict['all_loop'] == "1 1 1 2 2 2 "

def test_for_each_as_target_property_multidict(temp_factory):
    '''Test for_each option of input'''
    # test multi-key dictionary format of for_each
    temp_factory('a.txt')
    execute_workflow(r"""
        import itertools
        [0: shared=['counter', 'all_names', 'all_loop']]
        parameter: n = [300, 100]
        parameter: p = [50, 200, 100]
        parameter: outfile = ['1', '2', '3', '4', '5', '6']
        counter = 0
        all_names = ''
        all_loop = ''
        input: 'a.txt', group_by='single', for_each={'_n,_p': [(_n,_p) for _n,_p in itertools.product(n,p) if _n > _p]}

        all_names += outfile[_index] + " "
        all_loop += '{} {} '.format(_input._n, _input._p)
        counter = counter + 1
        """)
    assert env.sos_dict['counter'] == 4
    assert env.sos_dict['all_names'] == "1 2 3 4 "
    assert env.sos_dict['all_loop'] == "300 50 300 200 300 100 100 50 "

def test_for_each_as_target_property_nested_list(temp_factory):
    '''Test for_each option of input'''
    # test same-level for loop and parameter with nested list
    temp_factory('a.txt', 'b.txt', 'p1.txt', 'p2.txt', 'p3.txt')
    execute_workflow(
        r"""
        [0: shared=['processed']]
        files = ['a.txt', 'b.txt']
        processed = []

        input: files, for_each={'par':[(1, 2), (1, 3), (2, 3)], 'res': ['p1.txt', 'p2.txt', 'p3.txt']}
        output: _input.res

        processed.append((_input.par, _input.res))
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['processed'] == [((1, 2), 'p1.txt'), ((1, 3), 'p2.txt'),
                                         ((2, 3), 'p3.txt')]

def test_for_each_as_target_property_pandas(temp_factory):
    '''Test for_each option of input'''
    # test for each for pandas dataframe
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        input: for_each={'data': pd.DataFrame([(1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])}
        output: f"{_input.data['A']}_{_input.data['B']}_{_input.data['C']}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['1_2_Hello.txt', '2_4_World.txt']
    #

def test_for_each_as_target_property_index_type(temp_factory):
    '''Test for_each option of input'''
    # support for pands Series and Index types
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        data = pd.DataFrame([(1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])
        input: for_each={'A': data['A']}
        output: f"a_{_input.A}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['a_1.txt', 'a_2.txt']

def test_for_each_as_target_property_dataframe_keys(temp_factory):
    '''Test for_each option of input'''
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        data = pd.DataFrame([(1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])
        data.set_index('C', inplace=True)
        input: for_each={'A': data.index}
        output: f"{_input.A}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['Hello.txt', 'World.txt']

def test_for_each_as_target_property_series(temp_factory):
    '''Test for_each option of input'''
    # test for each of Series
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        data = pd.DataFrame([(0, 1, 'Ha'), (1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])

        data.set_index('A', inplace=True)
        data = data.tail(2)
        input: for_each={'A': data['B']}
        output: f"{_input.A}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['2.txt', '4.txt']

def test_for_each_as_target_property_iterables(temp_factory):
    '''Test for_each option of input'''
    # test iterable
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        import pandas as pd
        data = pd.DataFrame([(0, 1, 'Ha'), (1, 2, 'Hello'), (2, 4, 'World')], columns=['A', 'B', 'C'])

        data.set_index('A', inplace=True)
        data = data.tail(2)
        input: for_each={'A,B': zip(data['B'],data['C'])}
        output: f"{_input.A}.txt"
        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['res'] == ['2.txt', '4.txt']

def test_for_each_with_full_context(temp_factory):
    # test iterable
    execute_workflow(
        r"""
        input: for_each = [
            {'A': 1, 'B': 2},
            {'A': 5, 'B': 7}
            ]

        print(f'A={A}, B={B}')
        """)


def test_group_by_with_no_input():
    '''Test group_by with no input file'''
    execute_workflow(r'''
        [0]
        input: group_by=2
        ''')


def test_paired_with(temp_factory, clear_now_and_after):
    '''Test option paired_with '''
    temp_factory('a.txt', 'b.txt')
    clear_now_and_after('a.txt1', 'b.txt2')
    #
    # string input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        vars = [1, 2]

        input: files, paired_with='vars', group_by=1
        output: f"{_input}{_vars[0]}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1', 'b.txt2']:
        assert file_target(ofile).target_exists('target')
    #
    # list input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        vars = [1, 2]
        vars2 = ['a', 'b']

        input: files, paired_with=('vars', 'vars2'), group_by=1
        output: f"{_input}{_vars[0]}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1', 'b.txt2']:
        assert file_target(ofile).target_exists('target')
    #
    # dict input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        input: files, paired_with={'var': [1,2], 'var2': ['a', 'b']}, group_by=1
        output: f"{_input}{var[0]}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1', 'b.txt2']:
        assert file_target(ofile).target_exists('target')
    #
    # test paired_with paths
    clear_now_and_after('1.txt', '2.txt')
    execute_workflow(r'''
        [1]
        input: 'a.txt', 'b.txt', paired_with = dict(bgen=paths(['1.txt', '2.txt'])), group_by=1
        output: _input.bgen
        _output.touch()
        ''')
    for ofile in ['1.txt', '2.txt']:
        assert file_target(ofile).target_exists('target')


def test_paired_with_as_target_property(temp_factory):
    '''Test option paired_with with values accessed by individual target '''
    temp_factory('a.txt', 'b.txt')
    for ofile in ['a.txt1', 'b.txt2']:
        if file_target(ofile).exists():
            file_target(ofile).unlink()
    #
    # string input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        vars = [1, 2]

        input: files, paired_with='vars', group_by=1
        output: f"{_input}{_input._vars}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1', 'b.txt2']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()
    #
    # list input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        vars = [1, 2]
        vars2 = ['a', 'b']

        input: files, paired_with=('vars', 'vars2'), group_by=1
        output: f"{_input}{_input._vars}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1', 'b.txt2']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()
    #
    # dict input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        input: files, paired_with={'var': [1,2], 'var2': ['a', 'b']}, group_by=1
        output: f"{_input}{_input.var}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1', 'b.txt2']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()


def test_group_with(temp_factory):
    '''Test option group_with '''
    temp_factory('a.txt', 'b.txt')
    for ofile in ['a.txt1', 'b.txt2']:
        if file_target(ofile).exists():
            file_target(ofile).unlink()
    #
    # string input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        vars = [1, 2]

        input: files, group_with='vars', group_by=1
        output: f"{_input}{_vars}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1', 'b.txt2']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()
    #
    # list input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        vars = [1]
        vars2 = ['a']

        input: files, group_with=('vars', 'vars2'), group_by=2
        output: f"{_input[0]}{_vars}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()
    #
    # dict input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        input: files, group_with={'var': [1], 'var2': ['a']}, group_by=2
        output: f"{_input[0]}{var}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()


def test_output_group_with(temp_factory):
    '''Test option group_with in output statement'''
    temp_factory('a.txt', 'b.txt')
    for ofile in ['a.txt1', 'b.txt2']:
        if file_target(ofile).exists():
            file_target(ofile).unlink()
    #
    # string input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        vars = [1, 2]

        input: files, group_by=1
        output: f"{_input}.bak", group_with=dict(_vars=vars[_index])
        run: expand=True
            touch {_output}

        [1]
        assert(_vars == _index + 1)
        ''')
    for ofile in ['a.txt.bak', 'b.txt.bak']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()
    #
    # list input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        vars = [1]
        vars2 = ['a']

        input: files, group_by=2
        output: f"{_input[0]}1", group_with=('vars', 'vars2')
        run: expand=True
            touch {_output}

        [1]
        assert(_vars == 1)
        assert(_input._vars2 == 'a')
        ''')
    for ofile in ['a.txt1']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()
    #
    # dict input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        input: files, group_by=2
        output: f"{_input[0]}.bak",  group_with={'var': [1], 'var2': ['a']}
        run: expand=True
            touch {_output}

        [1]
        assert(var == 1)
        assert(var2 == 'a')
        ''')
    for ofile in ['a.txt.bak']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()


def test_group_with_as_target_property(temp_factory):
    '''Test option group_with '''
    temp_factory('a.txt', 'b.txt')
    for ofile in ['a.txt1', 'b.txt2']:
        if file_target(ofile).exists():
            file_target(ofile).unlink()
    #
    # string input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        vars = [1, 2]

        input: files, group_with='vars', group_by=1
        output: f"{_input}{_input._vars}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1', 'b.txt2']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()
    #
    # list input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        vars = [1]
        vars2 = ['a']

        input: files, group_with=('vars', 'vars2'), group_by=2
        output: f"{_input[0]}{_input._vars}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()
    #
    # dict input
    execute_workflow(r'''
        [0]
        files = ['a.txt', 'b.txt']
        input: files, group_with={'var': [1], 'var2': ['a']}, group_by=2
        output: f"{_input[0]}{_input.var}"
        run: expand=True
            touch {_output}
        ''')
    for ofile in ['a.txt1']:
        assert file_target(ofile).target_exists('target')
        file_target(ofile).unlink()


def test_input_pattern(temp_factory):
    '''Test option pattern of step input '''
    #env.verbosity = 4
    temp_factory('a-20.txt', 'b-10.txt')
    execute_workflow(
        r"""
        [0: shared=['base', 'name', 'par', '_output']]

        files = ['a-20.txt', 'b-10.txt']
        input: files, pattern=['{name}-{par}.txt', '{base}.txt']
        output: ['{}-{}-{}.txt'.format(x,y,z) for x,y,z in zip(_base, _name, _par)]

        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['base'] == ["a-20", 'b-10']
    assert env.sos_dict['name'] == ["a", 'b']
    assert env.sos_dict['par'] == ["20", '10']
    assert env.sos_dict['_output'] == ["a-20-a-20.txt", 'b-10-b-10.txt']


def test_input_pattern_as_target_property(temp_factory):
    '''Test option pattern of step input '''
    #env.verbosity = 4
    temp_factory('a-20.txt', 'b-10.txt')
    execute_workflow(
        r"""
        [0: shared=['_output']]

        files = ['a-20.txt', 'b-10.txt']
        input: files, pattern=['{name}-{par}.txt', '{base}.txt']
        output: [f'{x._base}-{x._name}-{x._par}.txt' for x in _input]

        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['_output'] == ["a-20-a-20.txt", 'b-10-b-10.txt']


def test_output_pattern(temp_factory):
    '''Test option pattern of step output'''
    #env.verbosity = 4
    temp_factory('a-20.txt', 'b-10.txt')
    execute_workflow(
        r"""
        [0: shared=['base', 'name', 'par', '_output']]

        files = ['a-20.txt', 'b-10.txt']
        input: files, pattern=['{name}-{par}.txt', '{base}.txt']
        output: expand_pattern('{base}-{name}-{par}.txt'), expand_pattern('{par}.txt')

        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['base'] == ["a-20", 'b-10']
    assert env.sos_dict['name'] == ["a", 'b']
    assert env.sos_dict['par'] == ["20", '10']
    assert env.sos_dict['_output'] == [
        'a-20-a-20.txt', 'b-10-b-10.txt', '20.txt', '10.txt'
    ]


def test_output_from_input(temp_factory):
    '''Test deriving output files from input files'''
    temp_factory('a.txt', 'b.txt')
    execute_workflow(
        r"""
        [0: shared={'counter':'counter', 'step':'step_output'}]
        files = ['a.txt', 'b.txt']
        counter = 0

        input: files, group_by='single'
        output: _input[0] + '.bak'
        _output.touch()
        counter += 1
        """,
        options={'run_mode': 'run'},
        config={'sig_mode': 'force'})
    assert env.sos_dict['counter'] == 2
    assert env.sos_dict['step'] == ['a.txt.bak', 'b.txt.bak']


def test_depends_from_input(temp_factory, clear_now_and_after):
    '''Test deriving dependent files from input files'''
    temp_factory('a.txt', 'b.txt')
    clear_now_and_after('a.txt.bak', 'b.txt.bak')
    execute_workflow(
        r"""
        [bak: provides='{file}.bak']
        _output.touch()

        [0: shared={'counter': 'counter', 'step':'step_depends'}]
        counter = 0

        input: 'a.txt', 'b.txt', group_by='single'
        depends: _input[0] + '.bak'
        counter += 1
        """,
        options={'run_mode': 'run'},
        config={'sig_mode': 'force'})
    assert env.sos_dict['counter'] == 2
    assert env.sos_dict['step'] == ['a.txt.bak', 'b.txt.bak']
    assert env.sos_dict['step'].groups == [['a.txt.bak'], ['b.txt.bak']]


def test_local_namespace(temp_factory):
    '''Test if steps are well separated.'''
    # interctive mode behave differently
    temp_factory('a.txt')
    script = textwrap.dedent(r"""
    [1]
    a = 1

    [2]
    # this should fail because a is defined in another step
    print(a)

    """)
    with pytest.raises(Exception):
        execute_workflow(script)
    # however, alias should be sent back
    execute_workflow(
        r"""
        [1: shared={'shared': 'step_output'}]
        input: 'a.txt'
        output: 'b.txt'

        [2: shared={'tt':'step_output'}]
        print(shared)

        output: [x + '.res' for x in shared]

        """,
        options={'run_mode': 'dryrun'})
    assert env.sos_dict['shared'] == ['b.txt']
    assert env.sos_dict['tt'] == ['b.txt.res']
    #
    # this include other variables set in the step
    execute_workflow(
        r"""
        [1: shared={'shared':'c', 'd':'d'}]
        input: 'a.txt'
        output: 'b.txt'

        c = 'c.txt'
        d = 1

        [2: shared={'d': 'e'}]
        # this should fail because a is defined in another step
        print(shared)

        output: shared

        e = d + 1

        """,
        options={'run_mode': 'dryrun'})
    # I would like to disallow accessing variables defined
    # in other cases.
    assert env.sos_dict['shared'] == 'c.txt'
    assert env.sos_dict['d'] == 2


def test_dynamic_output(temp_factory):
    '''Testing dynamic output'''
    #
    temp_factory(dir='temp')
    #
    execute_workflow('''
        [10: shared={'test':'step_output'}]
        ofiles = []
        output: dynamic(ofiles)

        for i in range(4):
            ff = 'temp/something{}.html'.format(i)
            ofiles.append(ff)
            with open(ff, 'w') as h:
                h.write('a')
        ''')
    assert env.sos_dict['test'] == [
        'temp/something{}.html'.format(x) for x in range(4)
    ]


def test_dynamic_input(temp_factory):
    '''Testing dynamic input'''
    #
    temp_factory(dir='temp')
    #
    script = SoS_Script('''
import os
[1]

from pathlib import Path
for i in range(5):
    Path(os.path.join('temp', f'test_{i}.txt')).touch()

[10: shared={'test':'step_output'}]
input: dynamic(os.path.join('temp', '*.txt')), group_by='single'
output: dynamic(os.path.join('temp', '*.txt.bak'))

run: expand=True
    touch {_input}.bak
''')
    wf = script.workflow()
    Base_Executor(wf).run()
    assert env.sos_dict['test'], sos_targets(
        [os.path.join('temp', 'test_{}.txt.bak'.format(x)) for x in range(5)]
    ) == f"Expecting {[os.path.join('temp', 'test_{}.txt.bak'.format(x)) for x in range(5)]} observed {env.sos_dict['test']}"
    # this time we use th existing signature
    Base_Executor(wf).run()
    assert env.sos_dict['test'], sos_targets(
        [os.path.join('temp', 'test_{}.txt.bak'.format(x)) for x in range(5)]
    ) == f"Expecting {[os.path.join('temp', 'test_{}.txt.bak'.format(x)) for x in range(5)]} observed {env.sos_dict['test']}"


def test_use_of_runmode(temp_factory):
    '''Test the use of run_mode variable in SoS script'''
    #
    temp_factory(dir='temp')
    env.config['sig_mode'] = 'ignore'
    execute_workflow('''
        [1: shared={'res': '_output'}]
        import random
        for i in range(3):
            with open(f"temp/test_{random.randint(1, 100000)}.txt", 'w') as res:
                res.write(str(i))

        ''')
    # we should have 9 files
    files = glob.glob(os.path.join('temp', '*.txt'))
    assert len(files) == 3


def test_action_before_input():
    '''Testing the execution of actions before input directive
    (variables such as _index should be made available). '''
    execute_workflow(
        '''
        [0]
        run('echo "A"')
        input:
        ''',
        options={'run_mode': 'dryrun'})


def test_duplicate_io_files(temp_factory):
    '''Test interpretation of duplicate input/output/depends'''
    temp_factory(dir='temp')
    # Test duplicate input
    os.system('touch temp/1.txt')
    execute_workflow('''
        [1]
        input: ['temp/1.txt' for x in range(5)]
        run: expand=True
        touch temp/{len(_input)}.input
                ''')
    assert os.path.isfile('temp/5.input')
    # Test duplicate output
    script = textwrap.dedent('''
    [1]
    output: ['temp/2.txt' for x in range(5)]
    run: expand=True
    touch temp/2.txt
    touch temp/{len(_output)}.output
    ''')
    with pytest.raises(Exception):
        execute_workflow(script)
    # Test duplicate depends
    script = textwrap.dedent('''
    [1]
    input: 'temp/1.txt'
    depends: ['temp/2.txt' for x in range(5)]
    output: 'temp/3.txt'
    run: expand=True
    touch temp/3.txt
    touch temp/{len(_depends)}.depends
    ''')
    with pytest.raises(Exception):
        execute_workflow(script)


def test_output_in_loop(temp_factory):
    '''Test behavior of {_output} when used in loop'''
    temp_factory(dir='temp')
    env.config['sig_mode'] = 'ignore'
    execute_workflow('''
        [default]
        s = [x for x in range(5)]
        output_files = ['temp/{}.txt'.format(x) for x in range(5)]
        input: for_each = ['s'], concurrent=False
        output: output_files[_index]
        run: active = 0
        rm -f temp/out.log
        run: expand=True
        echo {step_output} >> temp/out.log
        touch {step_output}
                ''')
    # output should have 1, 2, 3, 4, 5, respectively
    with open('temp/out.log') as out:
        assert len(out.read().split()) == 5
    #
    temp_factory(dir='temp')
    env.config['sig_mode'] = 'ignore'
    execute_workflow('''
    [default]
    s = [x for x in range(5)]
    output_files = ['temp/{}.txt'.format(x) for x in range(5)]
    input: for_each = ['s'], concurrent=False
    output: output_files[_index]
    run: active = 0
    rm -f temp/out.log
    run: expand=True
    echo {step_output} >> temp/out.log
    touch {step_output}
    ''')
    with open('temp/out.log') as out:
        assert len(out.read().split()) == 5


@multi_attempts
def test_execution_lock():
    '''Test execution lock of two processes'''
    with open('lock.sos', 'w') as lock:
        lock.write(r'''
import time
[A_1]
output: 'a.txt'
with open('a.txt', 'w') as txt:
txt.write('A1\n')

# A1 and A2 are independent
[A_2]
input: None
output: 'b.txt'
with open('b.txt', 'w') as txt:
txt.write('A2\n')
    ''')
    ret1 = subprocess.Popen('sos run lock -j1', shell=True)
    ret2 = subprocess.Popen('sos run lock -j1', shell=True)
    ret1.wait()
    ret2.wait()
    # two processes execute A_1 and A_2 separately, usually
    # takes less than 5 seconds
    file_target('lock.sos').unlink()


def test_removed_intermediate_files(clear_now_and_after):
    '''Test behavior of workflow with removed internediate files'''
    clear_now_and_after('a.txt', 'aa.txt')
    script = SoS_Script('''
[10]
output: 'a.txt'
run:
echo "a" > a.txt

[20]
output: 'aa.txt'
run: expand=True
cat {_input} > {_output}
''')
    wf = script.workflow()
    Base_Executor(wf).run()
    assert file_target('aa.txt').target_exists()
    # rerun should be faster
    Base_Executor(wf).run()
    # if we remove the middle result, it should not matter
    os.remove('a.txt')
    Base_Executor(wf).run()
    #
    # if we remove the final result, it will be rebuilt
    os.remove('aa.txt')
    Base_Executor(wf).run()
    #
    # now we request the generation of target
    file_target('a.txt').unlink()
    file_target('aa.txt').unlink()
    Base_Executor(wf).run()
    #
    file_target('a.txt').unlink()
    file_target('aa.txt').unlink()


def test_stopped_output():
    '''test output with stopped step'''
    for file in ["{}.txt".format(a) for a in range(10)]:
        if file_target(file).exists():
            file_target(file).unlink()
    execute_workflow('''
        [test_1]
        input: for_each={'a': range(10)}
        output: f"{a}.txt"

        stop_if(a % 2 == 0, no_output=True)
        run: expand=True
            touch {_output}

        [test_2]
        assert(len(step_input) == 5)
        ''')
    for idx in range(10):
        if idx % 2 == 0:
            assert not file_target("{}.txt".format(idx)).target_exists()
        else:
            assert file_target("{}.txt".format(idx)).target_exists()
            file_target(f"{idx}.txt").unlink()


def test_allow_error(clear_now_and_after):
    '''Test option allow error'''
    clear_now_and_after('a.txt')
    execute_workflow('''
        [test]
        run:  allow_error=True
            something_wrong

        run:
            touch a.txt
        ''')
    assert file_target('a.txt').target_exists()


def test_concurrent_worker(clear_now_and_after):
    '''Test the starting of multiple workers #493 '''
    with open('test_script.sos', 'w') as script:
        script.write('''
[10]
input: for_each={'i': range(1)}

[20]
input: for_each={'i': range(2)}
''')
    subprocess.call('sos run test_script', shell=True)
    clear_now_and_after('test_script.sos')


def test_depends_caused_dependency():
    # test for #674
    for tfile in ('1.txt', '2.txt', '3.txt'):
        if file_target(tfile).exists():
            file_target(tfile).unlink()
    execute_workflow('''
        [1: shared = {'dfile':'_output'}]
        output: '1.txt'
        run:
            echo 1 > 1.txt

        [2: shared = {'ifile':'_output'}]
        output: '2.txt'
        run: expand=True
            echo {_input} > 2.txt

        [3]
        depends: sos_variable('ifile'), sos_variable('dfile'), ifile
        input: dfile
        output: '3.txt'
        run: expand=True
            cat {_input} > {_output}
        ''')
    for tfile in ('1.txt', '2.txt', '3.txt'):
        assert file_target(tfile).target_exists()
        if file_target(tfile).exists():
            file_target(tfile).unlink()


def test_concurrent_input_option(temp_factory):
    '''Test input option'''
    temp_factory('1.txt', '2.txt')
    execute_workflow('''
        [1]
        n =[str(x) for x in range(2)]
        input: [f'{x+1}.txt' for x in range(2)], paired_with = 'n', concurrent = True
        run: expand = True
        echo {_n} {_input}
        ''')


def test_non_existent_dependent_target():
    '''Test non existent dependent targets'''
    with pytest.raises(Exception):
        execute_workflow(r"""
            [1]

            [2]
            depends: sos_step('wrong')
            """)
    with pytest.raises(Exception):
        execute_workflow(r"""
            [1]

            [2]
            depends: 'non-existent.txt'
            """)


def test_output_report(clear_now_and_after):
    '''Test generation of report'''
    clear_now_and_after('report.html')
    script = SoS_Script(r"""
[1: shared = {'dfile':'_output'}]
output: '1.txt'
run:
echo 1 > 1.txt

[2: shared = {'ifile':'_output'}]
output: '2.txt'
run: expand=True
echo {_input} > 2.txt

[3]
depends: ifile, sos_variable('ifile'), sos_variable('dfile')
input: dfile
output: '3.txt'
run: expand=True
cat {_input} > {_output}
""")
    env.config['output_report'] = 'report.html'
    wf = script.workflow()
    Base_Executor(wf).run()
    assert os.path.isfile('report.html')


@pytest.mark.skipif(
    sys.platform == 'win32', reason='Graphviz not available under windows')
def test_output_report_with_dag(clear_now_and_after):
    # test dag
    clear_now_and_after('report.html')
    execute_workflow(
        r"""
        [1: shared = {'dfile':'_output'}]
        output: '1.txt'
        run:
            echo 1 > 1.txt

        [2: shared = {'ifile':'_output'}]
        output: '2.txt'
        run: expand=True
            echo {_input} > 2.txt

        [3]
        depends: ifile, sos_variable('ifile'), sos_variable('dfile')
        input: dfile
        output: '4.txt'
        run: expand=True
            cat {_input} > {_output}
        """,
        options={
            'output_report': 'report.html',
            'output_dag': 'report.dag'
        })
    with open('report.html') as rep:
        content = rep.read()
    assert 'Execution DAG' in content


def test_sos_step_with_output():
    '''Test checking output of sos_step #981'''
    execute_workflow('''
        [step]
        output: 'a'
        sh:
        touch a

        [default]
        depends: sos_step('step')
        ''')


def test_multi_sos_step():
    '''Test matching 'a_1', 'a_2' etc with sos_step('a')'''
    for file in ('a_1', 'a_2'):
        if file_target(file).exists():
            file_target(file).unlink()
    res = execute_workflow('''
    [a_b_1]
    output: "a_1"
    sh:
    echo whatever > a_1

    [a_b_2]
    output: "a_2"
    sh: expand=True
    cp {_input} {_output}

    [default]
    depends: sos_step('a_b')
    ''')
    assert res['__completed__']['__step_completed__'] == 3
    assert os.path.isfile('a_1')
    assert os.path.isfile('a_2')
    with open('a_1') as a1, open('a_2') as a2:
        assert a1.read() == a2.read()


def test_depends_auxi_and_forward():
    '''Test depends on auxiliary, which then depends on a forward-workflow #983'''
    for f in ('a_1', 'a_2'):
        if file_target(f).exists():
            file_target(f).unlink()
    res = execute_workflow('''

    [hg_1]
    output: 'a_1'
    sh:
    echo "something" > a_1

    [hg_2]

    [star: provides = "a_2"]
    depends: sos_step('hg')
    sh:
    cp  a_1 a_2

    [default]
    depends: "a_2"
            ''')
    assert res['__completed__']['__step_completed__'] == 4
    assert os.path.isfile('a_1')
    assert os.path.isfile('a_2')
    with open('a_1') as a1, open('a_2') as a2:
        assert a1.read() == a2.read()


def test_depends_auxi_and_single_step_forward():
    '''Test depends on auxiliary, which then depends on a single-step forward-workflow'''
    for f in ('a_1', 'a_2'):
        if file_target(f).exists():
            file_target(f).unlink()
    res = execute_workflow('''

    [hg_1]
    output: 'a_1'
    sh:
    echo "something" > a_1

    [star: provides = "a_2"]
    depends: sos_step('hg')
    sh:
    cp  a_1 a_2

    [default]
    depends: "a_2"
    ''')
    assert res['__completed__']['__step_completed__'] == 3
    assert os.path.isfile('a_1')
    assert os.path.isfile('a_2')
    with open('a_1') as a1, open('a_2') as a2:
        assert a1.read() == a2.read()


def test_dryrun_placeholder():
    '''Test the creation and removal of placeholder files in dryrun mode'''
    if file_target('1.txt').exists():
        file_target('1.txt').unlink()
    execute_workflow(
        '''
        a = '1.txt'

        [out: provides=a]
        output: a
        run: expand = True
        touch {a}

        [1]
        depends: a
        ''',
        options={'run_mode': 'dryrun'})
    # but the file would be removed afterwards
    assert not os.path.isfile('1.txt')


def test_dryrun_in_sos_run(temp_factory):
    '''Test dryrun mode with sos_run #1007'''
    temp_factory('1.txt')
    script = SoS_Script(
        textwrap.dedent('''
        [remove]
        run:
        rm 1.txt

        [default]
        sos_run('remove')
        '''))
    wf = script.workflow()
    Base_Executor(wf).run(mode='dryrun')
    assert os.path.isfile('1.txt')
    Base_Executor(wf).run(mode='run')
    assert not os.path.isfile('1.txt')


def test_concurrent_with_dynamic_output(clear_now_and_after):
    '''Test concurrent steps with dynamic output'''
    douts = glob.glob('*.dout')
    for dout in douts:
        clear_now_and_after(dout)
    execute_workflow('''
        input: for_each={'i': range(3)}, concurrent=True
        output: dynamic('*.dout')
        import random
        path(f'{random.randint(0, 1000000)}.dout').touch()
        ''')
    douts = glob.glob('*.dout')
    assert len(douts) == 3


def test_group_by_with_emtpy_input():
    ''' Test option group by with empty input #1044'''
    execute_workflow('''
        [1]
        input: group_by=1
        print(_input)
        ''')


def test_removal_of_output_from_failed_step(clear_now_and_after):
    '''Test the removal of output files if a step fails #1055'''
    clear_now_and_after('failed.csv', 'result.csv')
    script = SoS_Script('''
[sub: provides='{file}.csv']
sh: expand=True
touch {_output}
eco "something wrong"

[step]
depends: 'failed.csv'
path('result.csv').touch()
''')
    wf = script.workflow()
    with pytest.raises(Exception):
        Base_Executor(wf).run()
    # rerun should still raise
    with pytest.raises(Exception):
        Base_Executor(wf).run()

    assert not os.path.isfile('failed.csv')
    assert not os.path.isfile('result.csv')


def test_depends_to_concurrent_substep():
    '''Testing forward style example'''
    # sos_variable('data') is passed to step [2]
    # but it is not passed to concurrent substep because
    # the variable is not used in the substep. This test
    # should fail at least under windows
    execute_workflow('''
        [1: shared={'data': 'step_output'}]
        output: 'a.txt'
        _output.touch()

        [2]
        depends: sos_variable('data')
        input: for_each={'i': range(2)}, group_by=1, concurrent=True
        print(1)
        ''')


def test_pass_of_target_source():
    '''Test passing of source information from step_output'''
    execute_workflow('''
        [1]
        output: 'a.txt'
        _output.touch()

        [2]
        assert step_input.labels == ['1']
        ''')
    #
    execute_workflow('''
        [1]
        input: for_each={'i': range(2)}
        output: 'a.txt', 'b.txt', group_by=1
        _output.touch()

        [2]
        assert step_input.labels == ['1', '1']
        ''')
    #
    file_target('c.txt').touch()
    execute_workflow('''
        [1]
        input: for_each={'i': range(2)}
        output: 'a.txt', 'b.txt', group_by=1
        _output.touch()

        [2]
        input: 'c.txt'
        assert step_input.labels == ['2']
        ''')


def test_rerun_with_zap():
    execute_workflow('''
        [step_10]
        input: for_each={'i': range(3)}
        output: f'zapped_example_{i}.txt'
        sh: expand=True
            echo "hello" > {_output}

        [step_20]
        input: group_by=1
        output: _input.with_suffix('.bak')
        sh: expand=True
            cp {_input} {_output}

        _input.zap()
        ''')
    execute_workflow('''
        [step_10]
        input: for_each={'i': range(3)}
        output: f'zapped_example_{i}.txt'
        sh: expand=True
            echo "hello" > {_output}

        [step_20]
        input: group_by=1
        output: _input.with_suffix('.bak')
        print(_output)
        sh: expand=True
            cp {_input} {_output}

        _input.zap()
        ''')
    for i in range(3):
        os.remove(f'zapped_example_{i}.txt.zapped')


def test_return_output_in_step_output():
    '''Testing the return of _output as groups of step_output'''
    execute_workflow('''\
        [1]
        input: for_each=dict(i=range(5))
        output: f'a_{i}.txt'
        _output.touch()
        assert(_input.i == i)

        [2]
        assert(len(step_input.groups) == 5)
        assert(len(step_input) == 5)
        assert(step_input.groups[0] == 'a_0.txt')
        assert(step_input.groups[4] == 'a_4.txt')
        #assert(_input.i == _index)
        ''')
    #
    # test accumulation of named output
    execute_workflow('''\
        [1]
        input: for_each=dict(i=range(5))
        output: a=f'a_{i}.txt', b=f'b_{i}.txt'
        _output.touch()

        [2]
        assert(len(step_input.groups) == 5)
        assert(len(step_input) == 10)
        assert(step_input.groups[0] == ['a_0.txt', 'b_0.txt'])
        assert(step_input.groups[0].labels == ['a', 'b'])
        assert(step_input.groups[4] == ['a_4.txt', 'b_4.txt'])
        assert(step_input.groups[4].labels == ['a', 'b'])
        ''')


def test_output_from():
    '''Testing output_from input function'''
    script = SoS_Script('''\
[A]
input: for_each=dict(i=range(5))
output: f'a_{i}.txt'
_output.touch()

[A1]
input: for_each=dict(i=range(4))
output: aa=f'a_{i}.txt'
_output.touch()

[B]
input: output_from('A')
assert(len(step_input.groups) == 5)
assert(len(step_input) == 5)
assert(step_input.labels == ['A']*5)
assert(step_input.groups[0] == 'a_0.txt')
assert(step_input.groups[4] == 'a_4.txt')

[C]
input: K=output_from('A')
assert(len(step_input.groups) == 5)
assert(step_input.labels == ['K']*5)

[D]
input: K=output_from('A', group_by='all')
assert(len(step_input) == 5)
assert(len(step_input.groups) == 1)
assert(step_input.labels == ['K']*5)

[E]
input: output_from('A1', group_by='all')
assert(len(step_input) == 4)
assert(len(step_input.groups) == 1)
assert(step_input.labels == ['aa']*4)

[F]
input: K=output_from('A1', group_by='all')['aa']
assert(len(step_input) == 4)
assert(len(step_input.groups) == 1)
assert(step_input.labels == ['K']*4)

[G_0]
input: for_each=dict(i=range(4))
output: f'g_{i}.txt'
_output.touch()

[G_100]
input: K=output_from(-1, group_by=2)
assert(len(step_input) == 4)
assert(len(step_input.groups) == 2)
assert(step_input.labels == ['K']*4)

[H_0]
input: for_each=dict(i=range(4))
output: f'g_{i}.txt'
_output.touch()

[H_100]
input: K=output_from([-1, 'A1'], group_by=2)
assert(len(step_input) == 8)
assert(len(step_input.groups) == 4)
assert(step_input.labels == ['K']*8)

''')
    for wf in ('B', 'C', 'D', 'E', 'F', 'G', 'H'):
        wf = script.workflow(wf)
        Base_Executor(wf).run()


def test_named_output1336():
    'Test issue 1336'
    execute_workflow('''
        import time

        [10]
        output: bak='A.bak'
        time.sleep(2)
        path('A.bak').touch()


        [20]
        input: named_output('bak')
        output: 'B.txt'
        _output.touch()
        ''')


def test_set_variables_to_output():
    '''Test assigning variables to _output'''
    execute_workflow('''\
        [10]
        output: 'a.txt'
        _output[0].set(tvar=1)
        _output.set(gvar=2)
        _output.touch()

        [20]
        assert(gvar == 2)
        assert(_input.gvar == 2)
        assert(_input.tvar == 1)
        assert(_input[0].tvar == 1)
        ''')
    # if there are substeps
    execute_workflow('''\
        [10]
        input: for_each=dict(i=range(4))
        output: f'a_{i}.txt'
        _output[0].set(tvar=i)
        _output.set(gvar=i)
        _output.touch()

        [20]
        assert(gvar == _index)
        assert(_input.gvar == _index)
        assert(_input.tvar == _index)
        assert(_input[0].tvar == _index)
        ''')


def test_step_id_vars():
    '''Test variables in a step'''
    execute_workflow('''
        [nested]
        print(f'Workflow {workflow_id}: step name={step_name}')
        print(f'Workflow {workflow_id}: step id={step_id}')
        print(f'Workflow {workflow_id}: workflow id={workflow_id}')
        print(f'Workflow {workflow_id}: master id={master_id}')
        assert step_name == 'nested'
        assert workflow_id != master_id

        [default]
        print(f'Workflow {workflow_id}: step name={step_name}')
        print(f'Workflow {workflow_id}: step id={step_id}')
        print(f'Workflow {workflow_id}: workflow id={workflow_id}')
        print(f'Workflow {workflow_id}: master id={master_id}')
        assert step_name == 'default'
        assert workflow_id == master_id
        sos_run('nested')
        ''')


def test_reexecution_of_dynamic_depends(clear_now_and_after):
    '''Testing the rerun of steps to verify dependency'''
    clear_now_and_after('a.bam', 'a.bam.bai')
    script = '''
        [BAI: provides='{filename}.bam.bai']
        _output.touch()

        [BAM]
        output: 'a.bam'
        _output.touch()

        [default]
        input: 'a.bam'
        depends: _input.with_suffix('.bam.bai')
        '''
    execute_workflow(script)
    # if we run again, because depends, the step will be re-checked
    os.remove('a.bam')
    res = execute_workflow(script)
    assert res['__completed__']['__step_completed__'] == 2
    assert res['__completed__']['__step_skipped__'] == 0
    #
    os.remove('a.bam')
    res = execute_workflow(script, options={'trace_existing': True})
    assert res['__completed__']['__step_completed__'] == 2
    assert res['__completed__']['__step_skipped__'] == 1


def test_traced_function(clear_now_and_after):
    clear_now_and_after('a.bam', 'a.bam.bai')
    execute_workflow('''
        [BAI: provides='{filename}.bam.bai']
        _output.touch()

        [BAM]
        output: 'a.bam'
        _output.touch()

        [default]
        input: 'a.bam'
        depends: traced(_input.with_suffix('.bam.bai'))
    ''')
    # if we run again, because depends, the step will be re-checked
    os.remove('a.bam')
    res = execute_workflow('''
        [BAI: provides='{filename}.bam.bai']
        _output.touch()

        [BAM]
        output: 'a.bam'
        _output.touch()

        [default]
        input: 'a.bam'
        depends: traced(_input.with_suffix('.bam.bai'))
        ''')
    assert res['__completed__']['__step_completed__'] == 2
    assert res['__completed__']['__step_skipped__'] == 1


def test_error_handling_of_step():
    # test fail_if of killing another running substep
    def cleanup():
        for step in (10, 11):
            if os.path.isfile(f'{step}.txt'):
                os.remove(f'{step}.txt')

    script = (
        textwrap.dedent(r"""
    import time

    [10]
    output: '10.txt'
    time.sleep(8)
    _output.touch()

    [11]
    output: '11.txt'
    _output.touch()

    [20]
    input: None
    time.sleep(2)
    fail_if(True)
    """))
    #
    # default mode
    #
    cleanup()
    st = time.time()
    with pytest.raises(Exception):
        execute_workflow(script)
    assert time.time(
    ) - st >= 8, 'Test test should fail only after step 10 is completed'
    assert os.path.isfile('10.txt')
    assert os.path.isfile('11.txt')
    #
    # ignore mode
    #
    cleanup()
    #
    st = time.time()
    execute_workflow(script, options={'error_mode': 'ignore'})
    assert time.time(
    ) - st >= 8, 'Test test should fail only after step 10 is completed'
    assert os.path.isfile('10.txt')
    assert os.path.isfile('11.txt')
    #
    # abort mode
    #
    cleanup()
    #
    with pytest.raises(Exception):
        execute_workflow(script, options={'error_mode': 'abort'})
    assert not os.path.isfile('10.txt')
    assert not os.path.isfile('11.txt')


def test_error_handling_of_substeps(clear_now_and_after):
    clear_now_and_after([f'test_{i}.txt' for i in range(10)],
                        [f'test_{i}.bak' for i in range(10)], 'test_30.txt')

    script = r"""
import time

[10]
input: for_each=dict(i=range(10)), concurrent=False
output: f'test_{i}.txt'

_output.touch()

fail_if(_index == 5, 'fail at 5')

[20]
output: _input.with_suffix('.bak')
_output.touch()

[30]
input: None
output: 'test_30.txt'
time.sleep(2)
_output.touch()
"""

    #
    # default mode
    #
    with pytest.raises(Exception):
        execute_workflow(script)

    for i in range(10):
        if i == 5:
            assert not os.path.isfile(f'test_{i}.txt')
            assert not os.path.isfile(f'test_{i}.bak')
        else:
            assert os.path.isfile(f'test_{i}.txt')
            # the following step is not executed
            assert not os.path.isfile(f'test_{i}.bak')
    # but the another branch continues
    assert os.path.isfile('test_30.txt')
    #
    # ignore mode
    #
    clear_now_and_after([f'test_{i}.txt' for i in range(10)],
                        [f'test_{i}.bak' for i in range(10)], 'test_30.txt')

    execute_workflow(script, options={'error_mode': 'ignore'})
    for i in range(6, 10):
        if i == 5:
            assert not os.path.isfile(f'test_{i}.txt')
            assert not os.path.isfile(f'test_{i}.bak')
        else:
            assert os.path.isfile(f'test_{i}.txt')
            assert os.path.isfile(f'test_{i}.bak')
    assert os.path.isfile('test_30.txt')
    #
    # abort mode
    #
    clear_now_and_after([f'test_{i}.txt' for i in range(10)],
                        [f'test_{i}.bak' for i in range(10)], 'test_30.txt')

    with pytest.raises(Exception):
        execute_workflow(script, options={'error_mode': 'abort'})
    for i in range(10):
        if i < 5:
            assert os.path.isfile(f'test_{i}.txt')
            assert not os.path.isfile(f'test_{i}.bak')
        else:
            assert not os.path.isfile(f'test_{i}.txt')
            assert not os.path.isfile(f'test_{i}.bak')
    assert not os.path.isfile('test_30.txt')


def test_error_handling_of_concurrent_substeps():

    def cleanup():
        for i in range(200):
            if os.path.isfile(f'test_{i}.txt'):
                os.remove(f'test_{i}.txt')

    script = textwrap.dedent(r"""
    import time

    [10]
    input: for_each=dict(i=range(200))
    output: f'test_{i}.txt'

    _output.touch()

    fail_if(_index == 5, 'fail at 5')
    fail_if(_index == 10, 'fail at 10')
    """)
    #
    # default mode
    #
    cleanup()
    with pytest.raises(Exception):
        execute_workflow(script)
    for i in (5, 10):
        assert not os.path.isfile(f'test_{i}.txt')
    for i in range(190, 200):
        assert os.path.isfile(f'test_{i}.txt')


def test_error_handling_of_tasks(clear_now_and_after):
    clear_now_and_after([f'test_{i}.txt' for i in (0, 1, 2, 30, 31)],
                        [f'test_{i}.bak' for i in (0, 1, 2, 30, 31)])

    script = r"""
import time

[10]
input: for_each=dict(i=range(3)), concurrent=False
output: f'test_{i}.txt'

task: queue='localhost'
fail_if(i == 1, 'fail at 1')
_output.touch()

[20]
output: _input.with_suffix('.bak')
_output.touch()

[30]
input: None
output: 'test_30.txt'
time.sleep(18)
_output.touch()

[31]
output: 'test_31.txt'
_output.touch()
    """
    #
    # default mode
    #

    with pytest.raises(Exception):
        execute_workflow(script)

    for i in (0, 2, 30, 31):
        assert os.path.isfile(f'test_{i}.txt')
        assert not os.path.isfile(f'test_{i}.bak')
    assert not os.path.isfile('test_1.txt')
    #
    #  ignore mode
    #
    clear_now_and_after([f'test_{i}.txt' for i in (0, 1, 2, 30, 31)],
                        [f'test_{i}.bak' for i in (0, 1, 2, 30, 31)])

    execute_workflow(script, options={'error_mode': 'ignore'})
    for i in (0, 2, 30, 31):
        assert os.path.isfile(f'test_{i}.txt')
    for i in (0, 2):
        assert not os.path.isfile(f'test_{i}.bak')
    assert not os.path.isfile('test_1.txt')
    #
    # abort mode
    #
    clear_now_and_after([f'test_{i}.txt' for i in (0, 1, 2, 30, 31)],
                        [f'test_{i}.bak' for i in (0, 1, 2, 30, 31)])

    with pytest.raises(Exception):
        execute_workflow(script, options={'error_mode': 'abort'})
    for i in (0, 2):
        assert os.path.isfile(f'test_{i}.txt')
    for i in (30, 31):
        assert not os.path.isfile(f'test_{i}.txt')
    for i in (0, 2):
        assert not os.path.isfile(f'test_{i}.bak')
    assert not os.path.isfile('test_1.txt')


def test_stmt_before_input(clear_now_and_after):
    '''Bug #1270, if there is any statement before input, the step will be undetermined'''

    clear_now_and_after('test_1270.txt', 'test_1270.out')
    execute_workflow(r'''
        [10]
        with open('test_1270.txt', 'w') as t1:
            t1.write('something')

        input: 'test_1270.txt'
        output: 'test_1270.out'
        with open(_input, 'r') as ifile, open(_output, 'w') as ofile:
            ofile.write(ifile.read())

        ''')

    assert os.path.isfile('test_1270.txt')
    assert os.path.isfile('test_1270.out')


def test_param_with_step_no_statement():
    # 1375
    execute_workflow(
        r'''
        [global]
        parameter: num = [x+1 for x in range(5)]

        [1]
        input: for_each = 'num'
        output: f'{_num}.txt'
        bash: expand = True
        touch {_output}

        [2]
        parameter: a = 1
        output: f'{_input:n}.out'
        task: trunk_workers = 1, trunk_size = 1, walltime = '3m', mem = '1G', cores = 1
        bash: expand = True
        echo {a} > {_output}
            ''',
        options={'default_queue': 'localhost'})


def test_parallel_nestedworkflow():
    # 1375
    execute_workflow(r'''
        [global]
        parameter: num = [x+1 for x in range(5)]

        [1]
        input: for_each = 'num'
        output: f'{_num}.txt'
        bash: expand = True
        touch {_output}

        [2]
        parameter: a = 1
        output: f'{_input:n}.out'
        sos_run('a' if a >=1 else 'b')

        [a,b]
        output: f'{_input:n}.out'
        bash: expand = True
        touch {_output}
        ''')


def test_concurrent_running_tasks(script_factory):
    '''Test two sos instances running the same task'''
    script = script_factory('''
        [1]
        task:
        import time
        time.sleep(5)
        ''')

    ret1 = subprocess.Popen(
        ['sos', 'run', script, '-s', 'force', '-q', 'localhost'])
    ret2 = subprocess.Popen(
        ['sos', 'run', script, '-s', 'force', '-q', 'localhost'])
    ret1.wait()
    ret2.wait()
    assert ret1.returncode == 0
    assert ret2.returncode == 0


def test_interpolation():
    '''Test string interpolation during execution'''
    execute_workflow(r"""
        [0: shared='res']
        res = ''
        b = 200
        res += f"{b}"
    """)
    assert env.sos_dict['res'] == '200'


def test_interpolation_1():
    '''Test string interpolation during execution'''
    execute_workflow(r"""
        [0: shared='res']
        res = ''
        for b in range(5):
            res += f"{b}"
    """)
    assert env.sos_dict['res'] == '01234'


def test_interpolation_2(temp_factory):
    '''Test string interpolation during execution'''
    temp_factory('a_1.txt', 'b_2.txt', 'c_2.txt')

    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        input: 'a_1.txt', 'b_2.txt', 'c_2.txt', pattern='{name}_{model}.txt'
        output: [f'{x}_{y}_processed.txt' for x,y in zip(name, model)]
        """,
        options={'run_mode': 'dryrun'})

    assert env.sos_dict['res'] == [
        'a_1_processed.txt', 'b_2_processed.txt', 'c_2_processed.txt'
    ]


def test_interpolation_3(temp_factory):
    '''Test string interpolation during execution'''
    temp_factory('a_1.txt', 'b_2.txt', 'c_2.txt')

    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        input: 'a_1.txt', 'b_2.txt', 'c_2.txt', pattern='{name}_{model}.txt'
        output: [f"{x}_{y}_process.txt" for x,y in zip(name, model)]
        """,
        options={'run_mode': 'dryrun'})

    assert env.sos_dict['res'] == [
        'a_1_process.txt', 'b_2_process.txt', 'c_2_process.txt'
    ]


def test_interpolation_4(temp_factory):
    '''Test string interpolation during execution'''
    temp_factory('a_1.txt', 'b_2.txt', 'c_2.txt')
    execute_workflow(
        r"""
        [0: shared={'res':'step_output'}]
        def add_a(x):
            return ['a'+_x for _x in x]

        input: 'a_1.txt', 'b_2.txt', 'c_2.txt', pattern='{name}_{model}.txt'
        output: add_a([f"{x}_{y}_process.txt" for x,y in zip(name, model)])

        """,
        options={'run_mode': 'dryrun'})

    assert env.sos_dict['res'] == [
        'aa_1_process.txt', 'ab_2_process.txt', 'ac_2_process.txt'
    ]


def test_named_output():
    '''Testing named_output input function'''
    for wf in ('B', 'C', 'D'):
        execute_workflow(
            '''
            [A]
            input: for_each=dict(i=range(4))
            output: aa=f'a_{i}.txt', bb=f'b_{i}.txt'
            _output.touch()

            [B]
            input: named_output('aa')
            assert(len(step_input.groups) == 4)
            assert(len(step_input) == 4)
            assert(step_input.labels == ['aa']*4)
            assert(step_input.groups[0] == 'a_0.txt')
            assert(step_input.groups[3] == 'a_3.txt')

            [C]
            input: K=named_output('bb')
            assert(len(step_input.groups) == 4)
            assert(len(step_input) == 4)
            assert(step_input.labels == ['K']*4)
            assert(step_input.groups[0] == 'b_0.txt')
            assert(step_input.groups[3] == 'b_3.txt')

            [D]
            input: K=named_output('bb', group_by=2)
            assert(len(step_input.groups) == 2)
            assert(len(step_input) == 4)
            assert(step_input.labels == ['K']*4)
            assert(step_input.groups[1] == ['b_2.txt', 'b_3.txt'])
            ''',
            workflow=wf)


def test_auto_provide():
    '''Testing steps to provide plain output'''
    execute_workflow('''
        [global]

        a = 'a.txt'

        [b]
        output: a
        _output.touch()

        [default]

        input: a
        output: f"{file_target(a):n}.out"

        _output.touch()
        ''')


def test_error_handling_of_missing_input(clear_now_and_after):
    # finishing another step if currently missing
    clear_now_and_after('11.txt', '22.txt')

    st = time.time()

    script = """
        import time

        [10]
        time.sleep(8)

        [11]
        output: '11.txt'
        _output.touch()

        [20]
        input: None
        time.sleep(2)

        [21]
        input: 'no_existent.txt'

        [22]
        output: '22.txt'
        _output.touch()
        """
    with pytest.raises(Exception):
        execute_workflow(script)

    assert time.time(
    ) - st >= 8, 'Test test should fail only after step 10 is completed'
    assert os.path.isfile('11.txt')

    clear_now_and_after('11.txt', '22.txt')

    execute_workflow(script, options={'error_mode': 'ignore'})

    assert time.time(
    ) - st >= 8, 'Test test should fail only after step 10 is completed'
    assert not os.path.isfile('22.txt')


def test_assignment_after_input(temp_factory):
    '''Testing assignment after input should be usable inside step process.'''
    #
    temp_factory(dir='temp')

    execute_workflow(
        '''
        [1]
        rep = range(5)
        input:  for_each='rep'
        output: f"temp/{_rep}.txt"

        # ff should change and be usable inside run
        ff = f"{_rep}.txt"
        run: expand=True
        echo {ff}
        touch temp/{ff}
        ''',
        options={'sig_mode': 'ignore'})


def test_1379(clear_now_and_after):
    clear_now_and_after('a.txt', 'b.txt')

    execute_workflow('''
        [multi: provides=['a.txt', 'b.txt']]

        out = 'a.txt', 'b.txt'
        output: out

        import time
        time.sleep(2)
        _output.touch()


        [step_1]
        input: 'a.txt'

        [step_2]
        input: 'b.txt'
        ''')


def test_remove_empty_groups_default():
    '''Test remove of empty groups'''
    # case 1, default output
    execute_workflow('''
        [10]
        input: for_each=dict(i=range(4))
        output: f'a_{i}.txt'
        _output.touch()
        skip_if(i==2)

        [20]
        assert len(step_input.groups) == 4
        ''')


def test_remove_empty_groups_false():
    '''Test remove of empty groups'''
    # case 2, use default remove_empty_groups=False
    execute_workflow('''
        [A]
        input: for_each=dict(i=range(4))
        output: f'a_{i}.txt'
        _output.touch()
        skip_if(i==2)

        [default]
        input: output_from('A')
        assert len(step_input.groups) == 4
        ''')


def test_remove_empty_groups_true():
    '''Test remove of empty groups'''
    # case 3, use remove_empty_groups=True
    execute_workflow('''
        [A]
        input: for_each=dict(i=range(4))
        output: f'a_{i}.txt'
        _output.touch()
        skip_if(i==2)

        [default]
        input: output_from('A', remove_empty_groups=True)
        assert len(step_input.groups) == 3
        ''')


def test_remove_empty_groups_named():
    '''Test remove of empty groups'''
    # case 4, use named_output
    execute_workflow('''
        [A]
        input: for_each=dict(i=range(4))
        output: A=f'a_{i}.txt'
        _output.touch()
        skip_if(i==2)

        [default]
        input: named_output('A')
        assert len(step_input.groups) == 4
        ''')


def test_remove_empty_groups_empty_named():
    '''Test remove of empty groups'''
    # case 5, use named_output
    execute_workflow('''
        [A]
        input: for_each=dict(i=range(4))
        output: A=f'a_{i}.txt'
        _output.touch()
        skip_if(i==2)

        [default]
        input: named_output('A', remove_empty_groups=True)
        assert len(step_input.groups) == 3
        ''')


def test_multi_depends(clear_now_and_after, temp_factory):
    '''Test a step with multiple depdendend steps'''

    clear_now_and_after('dbsnp.vcf', 'hg19.fa', 'f1.fastq', 'f2.fastq',
                        'f1.bam', 'f2.bam', 'f1.bam.idx', 'f2.bam.idx')
    temp_factory('f1.fastq', 'f2.fastq')

    execute_workflow(
        '''
        import time

        [refseq: provides='hg19.fa']
        time.sleep(1)
        _output.touch()

        [dbsnp: provides='dbsnp.vcf']
        _output.touch()

        [align_10]
        depends: 'hg19.fa'
        input: 'f1.fastq', 'f2.fastq', group_by=1, concurrent=True
        output: _input.with_suffix('.bam')
        _output.touch()

        [align_20]
        input: group_by=1, concurrent=True
        output: _input.with_suffix('.bam.idx')
        _output.touch()

        [call_10]
        depends: 'dbsnp.vcf', 'hg19.fa'

        [call_20]
        ''',
        workflow='align+call')

    for file in ('dbsnp.vcf', 'hg19.fa', 'f1.bam', 'f2.bam', 'f1.bam.idx',
                 'f2.bam.idx'):
        assert os.path.isfile(file)


def test_execute_ipynb():
    '''Test extracting and executing workflow from .ipynb files'''
    script = SoS_Script(filename='sample_workflow.ipynb')
    wf = script.workflow()
    Base_Executor(wf).run()


@pytest.mark.skipif(
    sys.platform == 'win32',
    reason='Skip test because travis fails on this test for unknown reason, also due to a bug in psutil under windows'
)
def test_kill_worker(script_factory):
    '''Test if the workflow can error out after a worker is killed'''
    import psutil
    import time

    script_file = script_factory('''
        import time

        [1]
        time.sleep(4)

        [2]
        time.sleep(4)
        ''')
    ret = subprocess.Popen(['sos', 'run', script_file])
    proc = psutil.Process(ret.pid)
    while True:
        children = proc.children(recursive=True)
        if children:
            children[0].terminate()
            break
        time.sleep(0.1)
    ret.wait()

    ret = subprocess.Popen(['sos', 'run', script_file])
    proc = psutil.Process(ret.pid)
    while True:
        children = proc.children(recursive=True)
        if children:
            children[0].kill()
            break
        time.sleep(0.1)
    ret.wait()


def test_kill_substep_worker(script_factory):
    '''Test if the workflow can error out after a worker is killed'''
    import psutil
    import time

    script_file = script_factory('''
        import time

        [1]
        input: for_each=dict(i=range(4))
        time.sleep(2)
        ''')
    ret = subprocess.Popen(['sos', 'run', script_file, '-j3'])
    proc = psutil.Process(ret.pid)
    while True:
        children = proc.children(recursive=True)
        print(children)
        if children:
            children[-1].terminate()
            break
        time.sleep(0.1)
    ret.wait()

    ret = subprocess.Popen(['sos', 'run', script_file, '-j3'])
    proc = psutil.Process(ret.pid)
    while True:
        children = proc.children(recursive=True)
        if children:
            children[-1].kill()
            break
        time.sleep(0.1)
    ret.wait()


@pytest.mark.skipif(
    True, reason='This test needs to be improved to make it consistent')
def test_kill_task(script_factory):
    '''Test if the workflow can error out after a worker is killed'''
    subprocess.call(['sos', 'purge', '--all'])
    import psutil
    import time

    script_file = script_factory('''
        [1]
        task:
        import time
        time.sleep(10)
        ''')
    ret = subprocess.Popen(
        ['sos', 'run', script_file, '-s', 'force', '-q', 'localhost'])
    proc = psutil.Process(ret.pid)

    while True:
        children = proc.children(recursive=True)
        execute = [x for x in children if 'execute' in x.cmdline()]
        if len(execute) >= 1:
            # a bug: if the process is killed too quickly (the signal
            # function is not called), this will fail.
            time.sleep(1)
            execute[0].terminate()
            break
        time.sleep(0.1)
    ret.wait()
    assert ret.returncode != 0

    ret = subprocess.Popen(['sos', 'run', script_file, '-q', 'localhost'])
    proc = psutil.Process(ret.pid)
    while True:
        children = proc.children(recursive=True)
        execute = [x for x in children if 'execute' in x.cmdline()]
        if len(execute) >= 1:
            time.sleep(1)
            execute[0].kill()
            break
        time.sleep(0.1)
    ret.wait()
    assert ret.returncode != 0


@pytest.mark.skipif(
    True, reason='This test needs to be improved to make it consistent')
def test_restart_orphaned_tasks(script_factory):
    '''Test restarting orphaned tasks which displays as running at first.'''
    import psutil
    import time
    subprocess.call(['sos', 'purge', '--all'])

    script_file = script_factory('''
        [1]
        task:
        import time
        time.sleep(12)
        ''')
    ret = subprocess.Popen(
        ['sos', 'run', script_file, '-s', 'force', '-q', 'localhost'])
    proc = psutil.Process(ret.pid)

    while True:
        children = proc.children(recursive=True)
        execute = [x for x in children if 'execute' in x.cmdline()]
        if len(execute) >= 1:
            # a bug: if the process is killed too quickly (the signal
            # function is not called), this will fail.
            time.sleep(1)
            execute[0].kill()
            break
        time.sleep(0.1)
    proc.kill()
    #
    ret = subprocess.Popen(['sos', 'run', script_file, '-q', 'localhost'])
    ret.wait()
    assert ret.returncode == 0